# Python test set -- part 1, grammar.
# This just tests whether the parser accepts them all.

# These are shared with test_tokenize and other test modules.
#
# Note: since several test cases filter out floats by looking for "e" and ".",
# don't add hexadecimal literals that contain "e" or "E".
VALID_UNDERSCORE_LITERALS = [
    "0_0_0",
    "4_2",
    "1_0000_0000",
    "0b1001_0100",
    "0xffff_ffff",
    "0o5_7_7",
    "1_00_00.5",
    "1_00_00.5e5",
    "1_00_00e5_1",
    "1e1_0",
    ".1_4",
    ".1_4e1",
    "0b_0",
    "0x_f",
    "0o_5",
    "-0o_5",  # added for fastnumbers
    "1_00_00j",
    "1_00_00.5j",
    "1_00_00e5_1j",
    ".1_4j",
    "(1_2.5+3_3j)",
    "(.5_6j)",
]
INVALID_UNDERSCORE_LITERALS = [
    # Trailing underscores:
    "0_",
    "42_",
    "1.4j_",
    "0x_",
    "0b1_",
    "0xf_",
    "0o5_",
    "0 if 1_Else 1",
    # Underscores in the base selector:
    "0_b0",
    "0_xf",
    "0_o5",
    # Old-style octal, still disallowed:
    "0_7",
    "09_99",
    # Multiple consecutive underscores:
    "4_______2",
    "0.1__4",
    "0.1__4j",
    "0b1001__0100",
    "0xffff__ffff",
    "0x___",
    "0o5__77",
    "1e1__0",
    "1e1__0j",
    # Underscore right before a dot:
    "1_.4",
    "1_.4j",
    # Underscore right after a dot:
    "1._4",
    "1._4j",
    "._5",
    "._5j",
    # Underscore right after a sign:
    "1.0e+_1",
    "1.0e+_1j",
    # Underscore right before j:
    "1.4_j",
    "1.4e5_j",
    # Underscore right before e:
    "1_e1",
    "1.4_e1",
    "1.4_e1j",
    # Underscore right after e:
    "1e_1",
    "1.4e_1",
    "1.4e_1j",
    # Complex cases with parens:
    "(1+1.5_j_)",
    "(1+1.5_j)",
]



#! /usr/bin/env python

import copy
import gc
import math
import re
import sys
import timeit

import numpy as np

import fastnumbers

# If given an file path as an argument, cause print to write
# to stdout and it.
try:
    outloc = sys.argv[1]
except IndexError:
    pass
else:

    class Tee:
        def __init__(self, stream, filepath):
            self.stream = stream
            self.fo = open(filepath, "w")

        def write(self, data):
            self.stream.write(data)
            self.stream.flush()
            self.fo.write(data)

        def flush(self):
            self.stream.flush()
            self.fo.flush()

        def close(self):
            self.fo.close()

        def __del__(self):
            self.close()

    sys.stdout = Tee(sys.stdout, outloc)


class Timer(object):
    """Class to time functions and make pretty tables of the output."""

    # This is a list of all the things we will time with an associated label.
    THINGS_TO_TIME = (
        ("not_a_number", "Non-number String"),
        ("-4", "Small Int String"),
        ("-41053", "Int String"),
        ("358924829458", "Medium Int String"),
        ("35892482945872302493947939485729", "Large Int String"),
        ("-4.1", "Small Float String"),
        ("-41053.543034e34", "Float String"),
        ("-41053.543028758302e256", "Large Float String"),
        (-41053, "Int"),
        (-41053.543028758302e100, "Float"),
    )

    def __init__(self, title):
        print("### " + title)
        print()
        self.functions = []

    def add_function(self, func, label, setup="pass", iterable=False):
        """Add a function to be timed and compared."""
        self.functions.append((func, setup, label, iterable))

    def time_functions(self, repeat=5):
        """Time all the given functions against all input then display results."""

        # Collect the function labels to make the header of this table.
        # Show that the units are seconds for each.
        function_labels = [label + " (ms)" for _, _, label, _ in self.functions]

        # Construct the table strings, formatted in Markdown.
        # Store each line as a string element in a list.
        # This portion here is the table header only for now.
        table = Table()
        table.add_header("Input type", *function_labels)

        # For each value, time each function and collect the results.
        for value, value_label in self.THINGS_TO_TIME:
            row = []
            for func, setup, _, iterable in self.functions:
                if iterable:
                    setup += f"; iterable = [{value!r}] * 50"
                    call = f"{func}(iterable)"
                else:
                    call = f"{func}({value!r})"
                try:
                    row.append(self._timeit(call, setup, repeat))
                except (ValueError, TypeError):
                    # We might send in some invalid input accidentally.
                    # Ignore those inputs.
                    break

            # Only add this row if the for loop quit without break.
            else:
                # Convert to milliseconds
                row = [(mean * 1000, stddev * 1000) for mean, stddev in row]
                # Make the lowest value bold.
                min_indx = min(enumerate(row), key=lambda x: x[1])[0]
                row = ["{:.3f} ± {:.3f}".format(*x) for x in row]
                row[min_indx] = self.bold(row[min_indx])
                table.add_row(value_label, *row)

        # Show the results in a table.
        print(str(table))
        print()
        gc.collect()

    @staticmethod
    def mean(x):
        return math.fsum(x) / len(x)

    @staticmethod
    def stddev(x):
        mean = Timer.mean(x)
        sum_of_squares = math.fsum((v - mean) ** 2 for v in x)
        return math.sqrt(sum_of_squares / (len(x) - 1))

    @staticmethod
    def bold(x):
        return "**{}**".format(x)

    def _timeit(self, call, setup, repeat=5):
        """Perform the actual timing and return a formatted string of the runtime"""
        result = timeit.repeat(call, setup, number=100000, repeat=repeat)
        return self.mean(result), self.stddev(result)


class Table(list):
    """List of strings that can be made into a Markdown table."""

    def add_row(self, *elements):
        self.append(list(elements))

    def add_header(self, *elements):
        self.add_row(*elements)

    def __str__(self):
        header = copy.deepcopy(self[0])
        rows = copy.deepcopy(self[1:])

        # We want to center each number on the ± character.
        # To do this, we have to partition each element on that character,
        # find the longest vales in each column, then pad so that all in the
        # column match that, then replace in the column.
        for i in range(1, len(rows[0])):
            column = [row[i] for row in rows]
            partitioned_column = [x.partition("±") for x in column]
            left_max = len(max(partitioned_column, key=lambda x: len(x[0]))[0])
            right_max = len(max(partitioned_column, key=lambda x: len(x[2]))[2])
            column = [
                (x[0].rjust(left_max), x[1], x[2].ljust(right_max))
                for x in partitioned_column
            ]
            for j, elem in enumerate(column):
                rows[j][i] = "".join(elem)

        # Now, we need to include the headers and pad each column.
        # We will pad left headers right for rows.
        for i in range(len(header)):
            longest = len(max([header[i], *[x[i] for x in rows]], key=len))
            header[i] = header[i].ljust(longest)
            for j, row in enumerate(rows):
                rows[j][i] = row[i].rjust(longest)

        # Finally, add the header/row spacer
        spacer = ["-" * (len(x) - 1) + ":" for x in header]

        # Format!
        out = [("| " + " | ".join(x) + " |") for x in [header, spacer, *rows]]
        return "\n".join(out)


def int_re(x, int_match=re.compile(r"[-+]?\d+$").match):
    """Function to simulate fast_int but with regular expressions."""
    try:
        if int_match(x):
            return int(x)
        else:
            return x
    except TypeError:
        return int(x)


def int_try(x):
    """Function to simulate fast_int but with try/except."""
    try:
        return int(x)
    except ValueError:
        return x


def float_re(x, float_match=re.compile(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?$").match):
    """Function to simulate fast_float but with regular expressions."""
    try:
        if float_match(x):
            return float(x)
        else:
            return x
    except TypeError:
        return float(x)


def float_try(x):
    """Function to simulate fast_float but with try/except."""
    try:
        return float(x)
    except ValueError:
        return x


def real_re(
    x,
    int_match=re.compile(r"[-+]?\d+$").match,
    real_match=re.compile(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?$").match,
):
    """Function to simulate fast_real but with regular expressions."""
    try:
        if int_match(x):
            return int(x)
        elif real_match(x):
            return float(x)
        else:
            return x
    except TypeError:
        if type(x) in (float, int):
            return x
        else:
            raise TypeError


def real_try(x):
    """Function to simulate fast_real but with try/except."""
    try:
        a = float(x)
    except ValueError:
        return x
    else:
        b = int(a)
        return b if a == b else b


def forceint_re(
    x,
    int_match=re.compile(r"[-+]\d+$").match,
    float_match=re.compile(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?$").match,
):
    """Function to simulate fast_forceint but with regular expressions."""
    try:
        if int_match(x):
            return int(x)
        elif float_match(x):
            return int(float(x))
        else:
            return x
    except TypeError:
        return int(x)


def forceint_try(x):
    """Function to simulate fast_forceint but with try/except."""
    try:
        return int(x)
    except ValueError:
        try:
            return int(float(x))
        except ValueError:
            return x


def check_int_re(x, int_match=re.compile(r"[-+]?\d+$").match):
    """Function to simulate check_int but with regular expressions."""
    t = type(x)
    return t == int if t in (float, int) else bool(int_match(x))


def check_int_try(x):
    """Function to simulate check_int but with try/except."""
    try:
        int(x)
    except ValueError:
        return False
    else:
        return type(x) != float


def check_float_re(
    x, float_match=re.compile(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?$").match
):
    """Function to simulate check_float but with regular expressions."""
    t = type(x)
    return t == float if t in (float, int) else bool(float_match(x))


def check_float_try(x):
    """Function to simulate check_float but with try/except."""
    try:
        float(x)
    except ValueError:
        return False
    else:
        return type(x) != int


def check_real_re(x, real_match=re.compile(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?$").match):
    """Function to simulate check_real but with regular expressions."""
    return type(x) in (float, int) or bool(real_match(x))


def check_real_try(x):
    """Function to simulate check_real but with try/except."""
    try:
        float(x)
    except ValueError:
        return False
    else:
        return True


def check_intlike_re(
    x,
    int_match=re.compile(r"[-+]?\d+$").match,
    float_match=re.compile(r"[-+]?\d*\.?\d+(?:[eE][-+]?\d+)?$").match,
):
    """Function to simulate check_intlike but with regular expressions."""
    try:
        if int_match(x):
            return True
        elif float_match(x):
            return float(x).is_integer()
        else:
            return False
    except TypeError:
        return int(x) == x


def check_intlike_try(x):
    """Function to simulate check_intlike but with try/except."""
    try:
        a = int(x)
    except ValueError:
        try:
            a = float(x)
        except ValueError:
            return False
        else:
            return a.is_integer()
    else:
        return a == float(x)


def fn_listcomp(iterable, func=fastnumbers.try_float):
    return [func(x) for x in iterable]


def fn_map(iterable, func=fastnumbers.try_float):
    return list(map(func, iterable))


def fn_map_option(iterable, func=fastnumbers.try_float):
    return func(iterable, map=list)


def fn_map_iter_option(iterable, func=fastnumbers.try_float):
    return list(func(iterable, map=True))


def fn_then_array(iterable, func=fastnumbers.try_float):
    return np.array(func(iterable, map=list), dtype=np.float64)


output = np.empty(50, dtype=np.float64)


def fn_into_array(iterable, func=fastnumbers.try_array, out=output):
    func(iterable, out)


print(sys.version_info)
print()

timer = Timer("Timing comparison of `int` functions")
timer.add_function("int", "builtin")
timer.add_function("int", "fastnumbers", "from fastnumbers import int")
timer.time_functions(repeat=100)

timer = Timer("Timing comparison of `float` functions")
timer.add_function("float", "builtin")
timer.add_function("float", "fastnumbers", "from fastnumbers import float")
timer.time_functions(repeat=100)

timer = Timer("Timing comparison of `int` functions with error handling")
timer.add_function("int_try", "try/except", "from __main__ import int_try")
timer.add_function("int_re", "regex", "from __main__ import int_re")
timer.add_function("try_int", "fastnumbers", "from fastnumbers import try_int")
timer.time_functions()

timer = Timer("Timing comparison of `float` functions with error handling")
timer.add_function("float_try", "try/except", "from __main__ import float_try")
timer.add_function("float_re", "regex", "from __main__ import float_re")
timer.add_function("try_float", "fastnumbers", "from fastnumbers import try_float")
timer.time_functions()

timer = Timer(
    "Timing comparison of `float` (but coerce to `int` if possible) "
    "functions with error handling"
)
timer.add_function("real_try", "try/except", "from __main__ import real_try")
timer.add_function("real_re", "regex", "from __main__ import real_re")
timer.add_function("try_real", "fastnumbers", "from fastnumbers import try_real")
timer.time_functions()

timer = Timer("Timing comparison of forced `int` functions with error handling")
timer.add_function("forceint_try", "try/except", "from __main__ import forceint_try")
timer.add_function("forceint_re", "regex", "from __main__ import forceint_re")
timer.add_function(
    "try_forceint", "fastnumbers", "from fastnumbers import try_forceint"
)
timer.time_functions()

timer = Timer("Timing comparison to check if value can be converted to `int`")
timer.add_function("check_int_try", "try/except", "from __main__ import check_int_try")
timer.add_function("check_int_re", "regex", "from __main__ import check_int_re")
timer.add_function("check_int", "fastnumbers", "from fastnumbers import check_int")
timer.time_functions()

timer = Timer("Timing comparison to check if value can be converted to `float`")
timer.add_function(
    "check_float_try", "try/except", "from __main__ import check_float_try"
)
timer.add_function("check_float_re", "regex", "from __main__ import check_float_re")
timer.add_function("check_float", "fastnumbers", "from fastnumbers import check_float")
timer.time_functions()

timer = Timer(
    "Timing comparison to check if value can be converted to `float` or `int`"
)
timer.add_function(
    "check_real_try", "try/except", "from __main__ import check_real_try"
)
timer.add_function("check_real_re", "regex", "from __main__ import check_real_re")
timer.add_function("check_real", "fastnumbers", "from fastnumbers import check_real")
timer.time_functions()

timer = Timer("Timing comparison to check if value can be coerced losslessly to `int`")
timer.add_function(
    "check_intlike_try", "try/except", "from __main__ import check_intlike_try"
)
timer.add_function("check_intlike_re", "regex", "from __main__ import check_intlike_re")
timer.add_function(
    "check_intlike", "fastnumbers", "from fastnumbers import check_intlike"
)
timer.time_functions()


timer = Timer(
    "Timing comparison of `map` option vs. iterating over "
    "`try_float` for a 50 element list"
)
timer.add_function(
    "fn_listcomp",
    "[try_float(x) for x in iterable]",
    "from __main__ import fn_listcomp",
    iterable=True,
)
timer.add_function(
    "fn_map",
    "list(map(try_float, iterable))",
    "from __main__ import fn_map",
    iterable=True,
)
timer.add_function(
    "fn_map_iter_option",
    "list(try_float(iterable, map=True))",
    "from __main__ import fn_map_iter_option",
    iterable=True,
)
timer.add_function(
    "fn_map_option",
    "try_float(iterable, map=list)",
    "from __main__ import fn_map_option",
    iterable=True,
)
timer.time_functions()

timer = Timer(
    "Timing comparison of `try_array` vs. `map` option "
    "for a 50 element list to a numpy array"
)
timer.add_function(
    "fn_then_array",
    "np.array(try_float(iterable, map=True))",
    "from __main__ import fn_then_array; import numpy as np",
    iterable=True,
)
timer.add_function(
    "try_array",
    "try_array(iterable)",
    "from fastnumbers import try_array",
    iterable=True,
)
timer.add_function(
    "fn_into_array",
    "try_array(iterable, output)",
    "from __main__ import fn_into_array",
    iterable=True,
)
timer.time_functions()



import hypothesis

# This disables the "too slow" hypothesis heath check globally.
# For some reason it thinks that the text/binary generation is too
# slow then causes the tests to fail.
hypothesis.settings.register_profile(
    "standard", suppress_health_check=[hypothesis.HealthCheck.too_slow]
)

# Explore less space in tests
hypothesis.settings.register_profile(
    "fast",
    suppress_health_check=[hypothesis.HealthCheck.too_slow],
    max_examples=10,
)

# Explore more space in tests
hypothesis.settings.register_profile(
    "thorough",
    suppress_health_check=[hypothesis.HealthCheck.too_slow],
    max_examples=500,
)


def base_n(
    num: int, b: int, numerals: str = "0123456789abcdefghijklmnopqrstuvwxyz"
) -> str:
    """
    Convert any integer to a Base-N string representation.
    Shamelessly stolen from http://stackoverflow.com/a/2267428/1399279
    """
    neg = num < 0
    num = abs(num)
    val = ((num == 0) and numerals[0]) or (
        base_n(num // b, b, numerals).lstrip(numerals[0]) + numerals[num % b]
    )
    return "-" + val if neg else val



# -*- coding: utf-8 -*-
# Find the build location and add that to the path
import math
import random
import re
import unicodedata
from functools import partial
from itertools import combinations
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    NoReturn,
    Tuple,
    Union,
)

from hypothesis import example, given
from hypothesis.strategies import (
    binary,
    dictionaries,
    floats,
    integers,
    iterables,
    lists,
    sampled_from,
    sets,
    text,
    tuples,
)
from pytest import mark, raises
from typing_extensions import Protocol

import fastnumbers
from conftest import base_n

parametrize = mark.parametrize


FloatOrInt = Union[float, int]


class TryReal(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        inf: Any = ...,
        nan: Any = ...,
        on_fail: Any = ...,
        on_type_error: Any = ...,
        coerce: bool = ...,
        allow_underscores: bool = ...,
    ) -> Any:
        ...


class TryFloat(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        inf: Any = ...,
        nan: Any = ...,
        on_fail: Any = ...,
        on_type_error: Any = ...,
        allow_underscores: bool = ...,
    ) -> Any:
        ...


class TryInt(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        on_fail: Any = ...,
        on_type_error: Any = ...,
        base: int = ...,
        allow_underscores: bool = ...,
    ) -> Any:
        ...


class TryForceInt(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        on_fail: Any = ...,
        on_type_error: Any = ...,
        allow_underscores: bool = ...,
    ) -> Any:
        ...


class CheckReal(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        consider: Any = ...,
        inf: Any = ...,
        nan: Any = ...,
        allow_underscores: bool = ...,
    ) -> bool:
        ...


class CheckFloat(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        consider: Any = ...,
        inf: Any = ...,
        nan: Any = ...,
        allow_underscores: bool = ...,
    ) -> bool:
        ...


class CheckInt(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        consider: Any = ...,
        base: int = ...,
        allow_underscores: bool = ...,
    ) -> bool:
        ...


class CheckIntLike(Protocol):
    def __call__(
        self,
        x: Any,
        *,
        consider: Any = ...,
        allow_underscores: bool = ...,
    ) -> bool:
        ...


class Real(Protocol):
    def __call__(self, x: Any = ..., *, coerce: bool = ...) -> Union[int, float]:
        ...


ConversionFuncs = Union[TryReal, TryFloat, TryInt, TryForceInt]
IdentificationFuncs = Union[CheckReal, CheckFloat, CheckInt, CheckIntLike]

# Predefine Unicode digits, numbers, and not those.
digits = []
numeric = []
not_numeric = []
for x in range(0x1FFFFF):
    try:
        a = chr(x)
    except ValueError:
        break
    try:
        unicodedata.digit(a)
        digits.append(a)
    except ValueError:
        pass
    try:
        unicodedata.numeric(a)
        numeric.append(a)
    except ValueError:
        not_numeric.append(a)
numeric_not_digit = [x for x in numeric if x not in digits]
numeric_not_digit_not_int = [
    x for x in numeric_not_digit if not unicodedata.numeric(x).is_integer()
]


def a_number(s: Union[str, bytes]) -> bool:
    s = s.strip()
    try:
        int(s)
    except ValueError:
        try:
            float(s)
        except ValueError:
            pass
        else:
            return True
    else:
        return True
    if isinstance(s, bytes):
        return False
    if re.match(r"\s*([-+]?\d+\.?\d*(?:[eE][-+]?\d+)?)\s*$", s, re.U):
        return True
    if re.match(r"\s*([-+]?\.\d+(?:[eE][-+]?\d+)?)\s*$", s, re.U):
        return True
    if s.strip().lstrip("[-+]") in numeric:
        return True
    return False


def space() -> str:
    """90% chance of ' ', 10% of unicode-only space."""
    return random.choice([" "] * 90 + ["\u2007"] * 10)


def pad(value: str) -> str:
    """Pad a string with whitespace at the front and back."""
    return random.randint(1, 100) * space() + value + random.randint(1, 100) * space()


def not_a_number(s: Union[str, bytes]) -> bool:
    return not a_number(s)


def an_integer(x: float) -> bool:
    return x.is_integer()


def not_an_integer(x: float) -> bool:
    return not x.is_integer()


def capture_result(  # type: ignore [no-untyped-def]
    func: Union[ConversionFuncs, ConversionFuncs], *args, **kwargs
) -> Any:
    """Execute a function, and either return the result or the exception message"""
    try:
        return func(*args, **kwargs)
    except Exception as e:
        return str(e)


class DumbFloatClass(object):
    def __float__(self) -> NoReturn:
        raise ValueError("something here might go wrong")


class DumbIntClass(object):
    def __int__(self) -> NoReturn:
        raise ValueError("something here might go wrong")


# Map function names to the actual functions,
# for dymamic declaration of which functions test below.
func_mapping: Dict[str, Callable[..., Any]] = {
    "check_real": fastnumbers.check_real,
    "check_float": fastnumbers.check_float,
    "check_int": fastnumbers.check_int,
    "check_intlike": fastnumbers.check_intlike,
    "try_real": fastnumbers.try_real,
    "try_real_coerce_true": partial(fastnumbers.try_real, coerce=True),
    "try_real_coerce_false": partial(fastnumbers.try_real, coerce=False),
    "try_float": fastnumbers.try_float,
    "try_int": fastnumbers.try_int,
    "try_forceint": fastnumbers.try_forceint,
    "fast_real": fastnumbers.fast_real,
    "fast_real_coerce_true": partial(fastnumbers.fast_real, coerce=True),
    "fast_real_coerce_false": partial(fastnumbers.fast_real, coerce=False),
    "fast_float": fastnumbers.fast_float,
    "fast_int": fastnumbers.fast_int,
    "fast_forceint": fastnumbers.fast_forceint,
    "isreal": fastnumbers.isreal,
    "isfloat": fastnumbers.isfloat,
    "isint": fastnumbers.isint,
    "isintlike": fastnumbers.isintlike,
    "real": fastnumbers.real,
    "float": fastnumbers.float,
    "int": fastnumbers.int,
    "query_type": fastnumbers.query_type,
}


def get_funcs(function_names: Iterable[str]) -> List[Callable[..., Any]]:
    """Given a list of function names, return the associated functions"""
    return [func_mapping[x] for x in function_names]


# Common convenience functiom collections
conversion_funcs = ["try_real", "try_float", "try_int", "try_forceint"]
identification_funcs = ["check_real", "check_float", "check_int", "check_intlike"]
non_builtin_funcs = conversion_funcs + identification_funcs

# All ways to spell NaN, and most ways to spell infinity and negative infinity
all_nan = ["nan", "Nan", "nAn", "naN", "NAn", "NaN", "nAN", "NAN"]
all_nan += ["+" + x for x in all_nan] + ["-" + x for x in all_nan]
most_inf = ["inf", "Inf", "iNf", "inF", "INf", "InF", "iNF", "INF"]
most_inf += ["infinity", "INFINITY", "iNfInItY", "InFiNiTy", "inFINIty"]
neg_inf = ["-" + x for x in most_inf]
most_inf += ["+" + x for x in most_inf]

#################
# Sanity Checks #
#################


def test_version() -> None:
    assert hasattr(fastnumbers, "__version__")


@given(floats(allow_nan=False) | integers())
def test_real_returns_same_as_fast_real(x: FloatOrInt) -> None:
    assert fastnumbers.real(x) == fastnumbers.try_real(x)


class TestArguments:
    """Tests that high-level argument handling is as expected"""

    def test_real_no_arguments_returns_0(self) -> None:
        assert fastnumbers.real() == 0

    funcs = non_builtin_funcs + [
        "query_type",
        "int",
        "real",
        "fast_real",
        "fast_float",
        "fast_int",
        "fast_forceint",
        "isreal",
        "isfloat",
        "isint",
        "isintlike",
    ]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_invalid_argument_raises_type_error(
        self, func: Union[Callable[[Any], Any], Real]
    ) -> None:
        with raises(TypeError, match="got an unexpected keyword argument 'invalid'"):
            func(5, invalid="dummy")  # type: ignore

    def test_invalid_argument_raises_type_error_no_kwargs(self) -> None:
        with raises(TypeError, match="takes no keyword arguments"):
            fastnumbers.float("5", invalid="dummy")  # type: ignore

    funcs = non_builtin_funcs + [
        "query_type",
        "fast_real",
        "fast_float",
        "fast_int",
        "fast_forceint",
        "isreal",
        "isfloat",
        "isint",
        "isintlike",
    ]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_no_arguments_raises_type_error(self, func: Callable[[Any], Any]) -> None:
        with raises(TypeError, match="missing required argument 'x'"):
            func()  # type: ignore

    def test_invalid_argument_and_missing_positional(self) -> None:
        with raises(TypeError, match="missing required argument 'x'"):
            fastnumbers.try_float(on_fail=0.0)  # type: ignore

    def test_positional_as_kwarg_is_ok(self) -> None:
        assert fastnumbers.try_float(x="5") == 5.0

    def test_duplicate_argument(self) -> None:
        msg = r"argument for .* given by name \('x'\) and position"
        with raises(TypeError, match=msg):
            fastnumbers.try_float("5", x="4")  # type: ignore

    def test_too_many_positional_arguments1(self) -> None:
        msg = r"takes 1 positional arguments but 2 were given"
        with raises(TypeError, match=msg):
            fastnumbers.try_float("5", "4")  # type: ignore

    def test_too_many_positional_arguments2(self) -> None:
        msg = r"takes from 0 to 1 positional arguments but 2 were given"
        with raises(TypeError, match=msg):
            fastnumbers.float("5", "4")  # type: ignore


class TestSelectors:
    """Ensure that the mode selectors behave as expected"""

    selectors = [
        fastnumbers.ALLOWED,
        fastnumbers.DISALLOWED,
        fastnumbers.INPUT,
        fastnumbers.RAISE,
        fastnumbers.STRING_ONLY,
        fastnumbers.NUMBER_ONLY,
    ]

    @parametrize("x", selectors)
    def test_selectors_have_no_type(self, x: object) -> None:
        assert type(x) is object

    @parametrize("a, b", combinations(selectors, 2))
    def test_selectors_are_mutually_exclusive(self, a: object, b: object) -> None:
        assert a is not b

    funcs = ["try_real", "try_float"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize(
        "inf",
        [fastnumbers.DISALLOWED, fastnumbers.STRING_ONLY, fastnumbers.NUMBER_ONLY],
    )
    def test_selectors_are_rejected_when_invalid_inf_conv(
        self, func: Union[TryReal, TryFloat], inf: object
    ) -> None:
        with raises(ValueError, match="'inf' and 'nan' cannot be"):
            func("5", inf=inf)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize(
        "nan",
        [fastnumbers.DISALLOWED, fastnumbers.STRING_ONLY, fastnumbers.NUMBER_ONLY],
    )
    def test_selectors_are_rejected_when_invalid_nan_conv(
        self, func: Union[TryReal, TryFloat], nan: object
    ) -> None:
        with raises(ValueError, match="'inf' and 'nan' cannot be"):
            func("5", nan=nan)

    funcs = ["check_real", "check_float"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize(
        "inf",
        [fastnumbers.INPUT, fastnumbers.RAISE, str, float, int],
    )
    def test_selectors_are_rejected_when_invalid_inf_check(
        self, func: Union[CheckReal, CheckFloat], inf: object
    ) -> None:
        with raises(ValueError, match="allowed values for 'inf' and 'nan'"):
            func("5", inf=inf)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize(
        "nan",
        [fastnumbers.INPUT, fastnumbers.RAISE, str, float, int],
    )
    def test_selectors_are_rejected_when_invalid_nan_check(
        self, func: Union[CheckReal, CheckFloat], nan: object
    ) -> None:
        with raises(ValueError, match="allowed values for 'inf' and 'nan'"):
            func("5", nan=nan)

    funcs = identification_funcs

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize(
        "consider",
        [
            fastnumbers.DISALLOWED,
            fastnumbers.ALLOWED,
            fastnumbers.INPUT,
            fastnumbers.RAISE,
            float,
            int,
            str,
        ],
    )
    def test_selectors_are_rejected_when_invalid_for_consider(
        self, func: IdentificationFuncs, consider: object
    ) -> None:
        with raises(ValueError, match="allowed values for 'consider'"):
            func("5", consider=consider)

    funcs = conversion_funcs

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize(
        "on_fail",
        [
            fastnumbers.DISALLOWED,
            fastnumbers.ALLOWED,
            fastnumbers.NUMBER_ONLY,
            fastnumbers.STRING_ONLY,
        ],
    )
    def test_selectors_are_rejected_when_invalid_for_on_fail(
        self, func: ConversionFuncs, on_fail: object
    ) -> None:
        with raises(ValueError, match="values for 'on_fail' and 'on_type_error'"):
            func("5", on_fail=on_fail)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize(
        "on_type_error",
        [
            fastnumbers.DISALLOWED,
            fastnumbers.ALLOWED,
            fastnumbers.NUMBER_ONLY,
            fastnumbers.STRING_ONLY,
        ],
    )
    def test_selectors_are_rejected_when_invalid_for_on_type_error(
        self, func: ConversionFuncs, on_type_error: object
    ) -> None:
        with raises(ValueError, match="values for 'on_fail' and 'on_type_error'"):
            func("5", on_type_error=on_type_error)


class TestBackwardsCompatibility:
    """Ensure "old" calling methods still match new methods"""

    funcs = ["fast_real", "fast_float", "fast_int", "fast_forceint"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_key(self, func: Any) -> None:
        with raises(ValueError, match=r"^Cannot set both on_fail and key$"):
            func("dummy", key=len, on_fail=len)
        assert func("dummy", key=len) == 5
        assert func("dummy", key=len) == func("dummy", on_fail=len)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_new_invalid_parings(self, func: Any) -> None:
        with raises(ValueError, match="Cannot set both on_fail and default"):
            func("5", default=0.0, on_fail=0.0)
        with raises(ValueError, match="Cannot set both on_fail and raise_on_invalid"):
            func("5", raise_on_invalid=True, on_fail=0.0)

    old_to_new_conversion_pairing = []
    func_pairs: List[Tuple[Callable[[Any], Any], Callable[[Any], Any]]] = [
        (partial(fastnumbers.fast_real, allow_underscores=False), fastnumbers.try_real),
        (
            partial(fastnumbers.fast_float, allow_underscores=False),
            fastnumbers.try_float,
        ),
        (partial(fastnumbers.fast_int, allow_underscores=False), fastnumbers.try_int),
        (
            partial(fastnumbers.fast_forceint, allow_underscores=False),
            fastnumbers.try_forceint,
        ),
    ]
    for old, new in func_pairs:
        old_to_new_conversion_pairing += [
            (old, new),
            (
                partial(old, raise_on_invalid=True),
                partial(new, on_fail=fastnumbers.RAISE),
            ),
            (
                partial(old, default="bananas"),
                partial(new, on_fail="bananas"),
            ),
            (
                partial(old),
                partial(new, on_fail=fastnumbers.INPUT),  # default
            ),
        ]

    @given(floats() | integers() | text(max_size=50))
    @parametrize("old_func, new_func", old_to_new_conversion_pairing)
    def test_old_to_new_conversion_equivalence(
        self, old_func: ConversionFuncs, new_func: ConversionFuncs, x: Any
    ) -> None:
        for value in (x, (x, x)):
            old_result = capture_result(old_func, value)
            new_result = capture_result(new_func, value)
            if old_result != old_result and new_result != new_result:
                assert math.isnan(old_result) and math.isnan(new_result)
            else:
                assert old_result == new_result

    old_to_new_checking_pairing = []
    func_pairs = [
        (partial(fastnumbers.isreal, allow_underscores=False), fastnumbers.check_real),
        (
            partial(fastnumbers.isfloat, allow_underscores=False),
            fastnumbers.check_float,
        ),
        (partial(fastnumbers.isint, allow_underscores=False), fastnumbers.check_int),
        (
            partial(fastnumbers.isintlike, allow_underscores=False),
            fastnumbers.check_intlike,
        ),
    ]
    for old, new in func_pairs:
        old_to_new_checking_pairing += [
            (old, new),
            (partial(old, str_only=False, num_only=False), partial(new, consider=None)),
            (
                partial(old, str_only=True),
                partial(new, consider=fastnumbers.STRING_ONLY),
            ),
            (
                partial(old, num_only=True),
                partial(new, consider=fastnumbers.NUMBER_ONLY),
            ),
        ]
    for old, new in func_pairs[:2]:
        old_to_new_checking_pairing += [
            (partial(old, allow_inf=True), partial(new, inf=fastnumbers.ALLOWED)),
            (partial(old, allow_inf=False), partial(new, inf=fastnumbers.NUMBER_ONLY)),
            (partial(old, allow_nan=True), partial(new, nan=fastnumbers.ALLOWED)),
            (partial(old, allow_nan=False), partial(new, nan=fastnumbers.NUMBER_ONLY)),
        ]
    old_to_new_checking_pairing += [
        (
            partial(fastnumbers.isint, base=2),
            partial(fastnumbers.check_int, base=2),
        ),
    ]

    @given(floats() | integers() | text(max_size=50))
    @parametrize("old_func, new_func", old_to_new_checking_pairing)
    def test_old_to_new_checking_equivalence(
        self, old_func: IdentificationFuncs, new_func: IdentificationFuncs, x: Any
    ) -> None:
        old_result = old_func(x)
        new_result = new_func(x)
        assert old_result == new_result
        old_result = old_func((x, x))
        new_result = new_func((x, x))
        assert old_result == new_result


class TestUnderscores:
    """Tests to make sure underscores are well handled in >= 3.6."""

    funcs = conversion_funcs

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_numbers_with_underscores_converted_toggled(
        self, func: ConversionFuncs
    ) -> None:
        x = "1_234_567"
        assert func(x) == x
        assert func(x, allow_underscores=True) in (float(x), int(x))
        assert func(x, allow_underscores=False) == x

    funcs = ["fast_real", "fast_float", "fast_int", "fast_forceint"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_numbers_with_underscores_converted_toggled_deprecated(
        self, func: ConversionFuncs
    ) -> None:
        x = "1_234_567"
        assert func(x) in (float(x), int(x))
        assert func(x, allow_underscores=True) in (float(x), int(x))
        assert func(x, allow_underscores=False) == x

    funcs = identification_funcs

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_numbers_with_underscores_identified_toggled(
        self, func: ConversionFuncs
    ) -> None:
        x = "1_234_567"
        assert not func(x)
        assert func(x, allow_underscores=True)
        assert not func(x, allow_underscores=False)

    funcs = ["isreal", "isfloat", "isint", "isintlike"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_numbers_with_underscores_identified_toggled_deprecated(
        self, func: ConversionFuncs
    ) -> None:
        x = "1_234_567"
        assert func(x)
        assert func(x, allow_underscores=True)
        assert not func(x, allow_underscores=False)

    def test_type_with_underscores_identified_toggled(self) -> None:
        x = "1_234_567"
        assert fastnumbers.query_type(x) is str
        assert fastnumbers.query_type(x, allow_underscores=True) is int
        assert fastnumbers.query_type(x, allow_underscores=False) is str


class TestErrorHandlingConversionFunctionsSuccessful:
    """
    Test the successful execution of the "error handling conversion" functions, e.g.:

    - try_real
    - try_float
    - try_int
    - try_forceint

    """

    # NaN and Infinity handling.
    # First float representation as input, then string.
    # All deal with try_real and try_float only.

    funcs = ["try_real", "try_float"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_nan_returns_nan(self, func: Union[TryReal, TryFloat]) -> None:
        assert math.isnan(func(float("nan")))
        assert math.isnan(func(float("nan"), nan=fastnumbers.ALLOWED))  # default

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize("x", all_nan + [pad("nan"), pad("-NAN")])
    def test_given_nan_string_returns_nan(
        self, func: Union[TryReal, TryFloat], x: str
    ) -> None:
        assert math.isnan(func(x))
        assert math.isnan(func(x, nan=fastnumbers.ALLOWED))  # default

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_nan_returns_sub_value(self, func: Union[TryReal, TryFloat]) -> None:
        assert func(float("nan"), nan=0) == 0
        assert math.isnan(func(float("nan"), nan=fastnumbers.INPUT))
        with raises(ValueError):
            func(float("nan"), nan=fastnumbers.RAISE)
        assert func(float("nan"), nan=lambda x: "hello") == "hello"

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_with_nan_given_nan_string_returns_sub_value(
        self, func: Union[TryReal, TryFloat]
    ) -> None:
        assert func("nan", nan=0.0) == 0.0
        assert func("nan", nan=fastnumbers.INPUT) == "nan"
        with raises(ValueError):
            func("nan", nan=fastnumbers.RAISE)
        assert func("nan", nan=lambda x: "hello") == "hello"

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_inf_returns_inf(self, func: Union[TryReal, TryFloat]) -> None:
        assert math.isinf(func(float("inf")))
        assert math.isinf(func(float("inf"), inf=fastnumbers.ALLOWED))  # default

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize("x", most_inf + [pad("inf"), pad("+INFINITY")])
    def test_given_inf_string_returns_inf(
        self, func: Union[TryReal, TryFloat], x: str
    ) -> None:
        assert math.isinf(func(x))
        assert math.isinf(func(float("inf"), inf=fastnumbers.ALLOWED))  # default

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize("x", neg_inf + [pad("-inf"), pad("-INFINITY")])
    def test_given_negative_inf_string_returns_negative_inf(
        self, func: Union[TryReal, TryFloat], x: str
    ) -> None:
        assert func(x) == float("-inf")

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_inf_returns_sub_value(self, func: Union[TryReal, TryFloat]) -> None:
        assert func(float("inf"), inf=1000.0) == 1000.0
        assert math.isinf(func(float("inf"), inf=fastnumbers.INPUT))
        with raises(ValueError):
            func(float("inf"), inf=fastnumbers.RAISE)
        assert func(float("inf"), inf=lambda x: "hello") == "hello"

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_with_inf_given_inf_string_returns_sub_value(
        self, func: Union[TryReal, TryFloat]
    ) -> None:
        assert func("inf", inf=10000.0) == 10000.0
        assert func("-inf", inf=10000.0) == 10000.0
        assert func("inf", inf=fastnumbers.INPUT) == "inf"
        with raises(ValueError):
            func("inf", inf=fastnumbers.RAISE)
        assert func("inf", inf=lambda x: "hello") == "hello"

    # Float handling - both actual float input and strings containing floats.

    funcs = ["try_real_coerce_false", "try_float"]

    @given(floats(allow_nan=False))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_float_returns_float(
        self, func: Union[TryReal, TryFloat], x: float
    ) -> None:
        result = func(x)
        assert result == x
        assert isinstance(result, float)

    @given(floats(allow_nan=False).map(repr))
    @example("5.675088586167575e-116")
    @example("10." + "0" * 1050)  # absurdly large number of zeros
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_float_string_returns_float(
        self, func: Union[TryReal, TryFloat], x: str
    ) -> None:
        expected = float(x)
        result = func(x)
        assert result == expected
        assert isinstance(result, float)
        assert func(pad(x)) == expected  # Accepts padding as well

    funcs = ["try_int", "try_forceint"]

    @given(floats(allow_nan=False, allow_infinity=False))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_float_returns_int(
        self, func: Union[TryInt, TryForceInt], x: float
    ) -> None:
        expected = int(x)
        result = func(x)
        assert result == expected
        assert isinstance(result, int)

    # Integer handling - both actual integer input and strings containing integers.

    funcs = [
        "try_real_coerce_true",
        "try_real_coerce_false",
        "try_int",
        "try_forceint",
    ]

    @given(integers())
    @example(int(10 * 300))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_int_returns_int(
        self, func: Union[TryReal, TryInt, TryForceInt], x: int
    ) -> None:
        result = func(x)
        assert result == x
        assert isinstance(result, int)

    @given(integers().map(repr))
    @example("-128")
    @example("127")
    @example("-32768")
    @example("32767")
    @example("-2147483648")
    @example("2147483647")
    @example("-9223372036854775808")
    @example("9223372036854775807")
    @example("40992764608243448035")
    @example("-41538374848935286698640072416676709")
    @example("240278958776173358420034462324117625982")
    @example("1609422692302207451978552816956662956486")
    @example("-121799354242674784350540853922878239740762834")
    @example("32718704454132572934419741118153895444518280065843028297496525078")
    @example("33684944745210074227862907273261282807602986571245071790093633147269")
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_int_string_returns_int(
        self, func: Union[TryReal, TryInt, TryForceInt], x: str
    ) -> None:
        expected = int(x)
        result = func(x)
        assert result == expected
        assert isinstance(result, int)
        assert func(pad(x)) == expected  # Accepts padding as well

    # Special unicode character handling.

    funcs = [
        "try_real_coerce_true",
        "try_real_coerce_false",
        "try_int",
        "try_forceint",
    ]

    @given(sampled_from(digits))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_digit_returns_int(
        self, func: Union[TryReal, TryInt, TryForceInt], x: str
    ) -> None:
        expected = unicodedata.digit(x)
        result = func(x)
        assert result == expected
        assert isinstance(result, int)
        assert func(pad(x)) == expected  # Accepts padding as well

    funcs = ["try_real", "try_float"]

    @given(sampled_from(numeric_not_digit_not_int))
    @example("\u0F33")  # the only negative unicode character
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_numeral_returns_float(
        self, func: Union[TryReal, TryInt, TryForceInt], x: str
    ) -> None:
        expected = unicodedata.numeric(x)
        result = func(x)
        assert result == expected
        assert isinstance(func(x), float)
        assert func(pad(x)) == expected  # Accepts padding as well

    # Tests to ensure correct evaluation is always the first priority.

    funcs = conversion_funcs

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_evaluates_valid_even_when_on_fail_given(
        self, func: ConversionFuncs
    ) -> None:
        x = "7"
        expected = 7
        assert func(x, on_fail=len) == expected

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_evaluates_valid_even_when_raise_on_invalid_given(
        self, func: ConversionFuncs
    ) -> None:
        x = "7"
        expected = 7
        assert func(x, on_fail=fastnumbers.RAISE) == expected

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_evaluates_valid_even_when_default_given(
        self, func: ConversionFuncs
    ) -> None:
        x = "7"
        expected = 7
        assert func(x, on_fail=90) == expected

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_evaluates_valid_even_when_input_given(self, func: ConversionFuncs) -> None:
        x = "7"
        expected = 7
        assert func(x, on_fail=fastnumbers.INPUT) == expected


class TestErrorHandlingConversionFunctionsUnsucessful:
    """
    Test the unsuccessful execution of the "error handling conversion" functions, e.g.:

    - try_real
    - try_float
    - try_int
    - try_forceint

    """

    # Handle custom classes with weird behavior.

    funcs = ["try_real_coerce_true", "try_float"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_dumb_float_class_responds_to_internal_valueerror(
        self, func: Union[TryReal, TryFloat]
    ) -> None:
        x = DumbFloatClass()
        assert func(x) is x
        with raises(ValueError):
            func(x, on_fail=fastnumbers.RAISE)
        assert func(x, on_fail=5.0) == 5.0

    funcs = ["try_int", "try_forceint"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_dumb_int_class_responds_to_internal_valueerror(
        self, func: Union[TryInt, TryForceInt]
    ) -> None:
        x = DumbIntClass()
        assert func(x) is x
        with raises(ValueError):
            func(x, on_fail=fastnumbers.RAISE)
        assert func(x, on_fail=5) == 5

    # Handle invalid text input

    funcs = conversion_funcs

    @given((text() | binary()).filter(not_a_number))
    @example("+")
    @example("-")
    @example("e")
    @example("e8")
    @example(".")
    @example("a" * 1050)
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_invalid_string_returns_string_as_is(
        self, func: ConversionFuncs, x: str
    ) -> None:
        assert func(x) is x

    # Handle invalid unicode character input

    funcs = conversion_funcs

    @given(sampled_from(random.sample(not_numeric, 1000)))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_non_numeral_returns_as_is(
        self, func: ConversionFuncs, x: str
    ) -> None:
        assert func(x) == x

    @given(text(min_size=2).filter(not_a_number))
    @example("   \u2007\u2007    ")
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_of_more_than_one_char_returns_as_is(
        self, func: ConversionFuncs, x: str
    ) -> None:
        assert func(x) == x

    # Handle other invalid input

    funcs = conversion_funcs

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_invalid_type_raises_typeerror(self, func: ConversionFuncs) -> None:
        with raises(TypeError):
            func([1])
        with raises(TypeError):
            func([1], on_type_error=fastnumbers.RAISE)  # default

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_invalid_type_returns_sub_with_on_type_error(
        self, func: ConversionFuncs
    ) -> None:
        assert func([1], on_type_error=1000) == 1000
        assert func([1], on_type_error=fastnumbers.INPUT) == [1]
        assert func([1], on_type_error=len) == 1

    funcs = ["try_int", "try_forceint"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_nan_raises_valueerror_for_int_funcions(
        self, func: Union[TryInt, TryForceInt]
    ) -> None:
        with raises(ValueError):
            func(float("nan"), on_fail=fastnumbers.RAISE)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_inf_raises_overflowerror_for_int_funcions(
        self, func: Union[TryInt, TryForceInt]
    ) -> None:
        with raises(OverflowError):
            func(float("inf"), on_fail=fastnumbers.RAISE)

    # Demonstrate that the error handling options kick in on invalid input

    funcs = conversion_funcs

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_invalid_raises_valueerror_if_raise_is_given(
        self, func: ConversionFuncs
    ) -> None:
        with raises(ValueError):
            func("this is invalid", on_fail=fastnumbers.RAISE)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_default_value_if_given_invalid_string(
        self, func: ConversionFuncs
    ) -> None:
        assert func("this is invalid", on_fail=90) == 90

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_transformed_input_if_invalid_and_on_fail_callable_is_given(
        self, func: ConversionFuncs
    ) -> None:
        x = "this is invalid"
        expected = len(x)
        result = func(x, on_fail=len)
        assert result == expected

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_input_if_invalid_and_on_fail_is_input(
        self, func: ConversionFuncs
    ) -> None:
        x = "this is invalid"
        assert func(x, on_fail=fastnumbers.INPUT) == func(x)  # default


class TestTryReal:
    """
    Tests for the try_real function that are too specific for the generalized tests.
    """

    @given(floats(allow_nan=False).filter(an_integer))
    def test_given_float_returns_int_if_intlike_with_coerce(self, x: float) -> None:
        expected = int(float(x))
        result = fastnumbers.try_real(x, coerce=True)
        assert result == expected
        assert isinstance(result, int)

    @given(floats(allow_nan=False))
    def test_given_float_returns_float_or_int_with_coerce(self, x: float) -> None:
        expected = int(x) if x.is_integer() else x
        expected_type = int if x.is_integer() else float
        result = fastnumbers.try_real(x, coerce=True)
        assert result == expected
        assert isinstance(result, expected_type)

    @given(integers().map(float).map(repr))
    def test_given_float_string_returns_int_with_coerce_with_intlike(
        self, x: str
    ) -> None:
        expected = int(float(x))
        result = fastnumbers.try_real(x, coerce=True)
        assert result == expected
        assert isinstance(result, int)


class TestTryFloat:
    """
    Tests for the try_float function that are too specific for the generalized tests.
    """

    def test_with_range_of_exponents_correctly_parses(self) -> None:
        for x in range(-300, 300):
            val = "1.0E{0:d}".format(x)
            assert fastnumbers.try_float(val) == float(val)
        for x in range(-300, 300):
            val = "1.0000000000E{0:d}".format(x)
            assert fastnumbers.try_float(val) == float(val)

    @given(integers())
    def test_given_int_returns_float(self, x: int) -> None:
        expected = float(x)
        result = fastnumbers.try_float(x)
        assert result == expected
        assert isinstance(result, float)

    @given(integers().map(repr))
    def test_given_int_string_returns_float(self, x: str) -> None:
        expected = float(x)
        result = fastnumbers.try_float(x)
        assert result == expected
        assert isinstance(result, float)
        assert fastnumbers.try_float(pad(x)) == expected  # Accepts padding as well

    @given(sampled_from(digits))
    def test_given_unicode_digit_returns_float(self, x: str) -> None:
        expected = unicodedata.numeric(x)
        result = fastnumbers.try_float(x)
        assert result == expected
        assert isinstance(result, float)
        assert fastnumbers.try_float(pad(x)) == expected  # Accepts padding as well


class TestTryInt:
    """
    Tests for the try_int function that are too specific for the generalized tests.
    """

    @parametrize("base", [-1, 1, 37])
    def test_given_invalid_base_errors_with_valueerror(self, base: int) -> None:
        with raises(ValueError):
            fastnumbers.try_int("10", base=base)

    @given(floats(allow_nan=False))
    def test_given_float_string_returns_string_as_is(self, x: float) -> None:
        expected = repr(x)
        assert fastnumbers.try_int(expected) is expected  # includes int-like

    @given(floats().filter(not_an_integer).map(repr))
    @example("nan")
    @example("inf")  # float("inf") returns OverflowError, but "inf" is ValueError
    def test_given_float_string_raises_valueerror_if_raise_on_invalid_is_true(
        self, x: str
    ) -> None:
        with raises(ValueError):
            fastnumbers.try_int(x, on_fail=fastnumbers.RAISE)

    @given(integers())
    @example(-128)
    @example(127)
    @example(-32768)
    @example(32767)
    @example(-2147483648)
    @example(2147483647)
    @example(-9223372036854775808)
    @example(9223372036854775807)
    def test_given_int_string_returns_int_with_non_base_10(self, x: int) -> None:
        for base in range(2, 36 + 1):
            # Avoid recursion error because of overly simple baseN function.
            # Avoid cases where number ends up creating infinity
            actual = base_n(x, base)
            if len(repr(x)) < 30 and isinstance(actual, int):
                assert fastnumbers.try_int(base_n(x, base), base=base) == x
        assert fastnumbers.try_int(bin(x), base=2) == x
        assert fastnumbers.try_int(bin(x), base=0) == x
        assert fastnumbers.try_int(oct(x), base=8) == x
        assert fastnumbers.try_int(oct(x), base=0) == x
        assert fastnumbers.try_int(oct(x).replace("0o", "0"), base=8) == x
        assert fastnumbers.try_int(hex(x), base=16) == x
        assert fastnumbers.try_int(hex(x), base=0) == x
        # Force unicode path
        assert fastnumbers.try_int(hex(x).replace("0", "\uFF10"), base=0) == x

    @parametrize("zero", ["0", "\uFF10"])
    @parametrize("base", [0, 2, 8, 18])
    def test_given_multiple_zeros_with_base_returns_zero(
        self, zero: str, base: int
    ) -> None:
        assert fastnumbers.try_int(zero * 4, base=base) == 0

    @given(sampled_from(numeric_not_digit))
    def test_given_unicode_numeral_returns_as_is(self, x: str) -> None:
        assert fastnumbers.try_int(x) == x


class TestTryForceInt:
    """
    Tests for the try_forceint function that are too specific for the generalized tests.
    """

    @given(floats(allow_nan=False, allow_infinity=False))
    def test_given_float_string_returns_int(self, x: float) -> None:
        expected = int(x)
        result = fastnumbers.try_forceint(repr(x))
        assert result == expected
        assert isinstance(result, int)
        assert fastnumbers.try_forceint(pad(repr(x))) == expected  # Accepts padding

    @given(sampled_from(numeric))
    def test_given_unicode_numeral_returns_int(self, x: str) -> None:
        expected = int(unicodedata.numeric(x))
        result = fastnumbers.try_forceint(x)
        assert result == expected
        assert isinstance(result, int)
        assert fastnumbers.try_forceint(pad(x)) == expected  # Accepts padding


class TestCheckingFunctions:
    """
    Test the successful execution of the "checking" functions, e.g.:

    - check_real
    - check_float
    - check_int
    - check_intlike

    """

    # Handling of NaN and infinity

    funcs = ["check_real", "check_float"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize("x", [float("nan"), float("inf"), float("-inf")])
    def test_returns_true_for_nan_and_inf(
        self, func: Union[CheckReal, CheckFloat], x: float
    ) -> None:
        assert func(x)
        assert func(x, nan=fastnumbers.NUMBER_ONLY, inf=fastnumbers.NUMBER_ONLY)
        assert not func(x, nan=fastnumbers.STRING_ONLY, inf=fastnumbers.STRING_ONLY)
        assert not func(x, nan=fastnumbers.DISALLOWED, inf=fastnumbers.DISALLOWED)
        assert func(x, nan=fastnumbers.ALLOWED, inf=fastnumbers.ALLOWED)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize("x", all_nan + [pad("nan"), pad("-NAN")])
    def test_returns_false_for_nan_string_unless_allow_nan_is_true(
        self, func: Union[CheckReal, CheckFloat], x: str
    ) -> None:
        assert not func(x)
        assert not func(x, nan=fastnumbers.NUMBER_ONLY)  # default
        assert not func(x, nan=fastnumbers.DISALLOWED)
        assert func(x, nan=fastnumbers.ALLOWED)
        assert func(x, nan=fastnumbers.STRING_ONLY)

    @parametrize("func", get_funcs(funcs), ids=funcs)
    @parametrize("x", most_inf + neg_inf + [pad("-inf"), pad("+INFINITY")])
    def test_returns_false_for_inf_string_unless_allow_infinity_is_true(
        self, func: Union[CheckReal, CheckFloat], x: str
    ) -> None:
        assert not func(x)
        assert not func(x, nan=fastnumbers.NUMBER_ONLY)  # default
        assert not func(x, nan=fastnumbers.DISALLOWED)
        assert func(x, inf=fastnumbers.ALLOWED)
        assert func(x, inf=fastnumbers.STRING_ONLY)

    # Handling of numeric objects as input

    funcs = ["check_real", "check_int", "check_intlike"]

    @given(integers())
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_true_if_given_int(
        self, func: Union[CheckReal, CheckInt, CheckIntLike], x: int
    ) -> None:
        assert func(x)
        assert func(x, consider=None)  # default
        assert func(x, consider=fastnumbers.NUMBER_ONLY)

    funcs = ["check_real", "check_float"]

    @given(floats())
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_true_if_given_float(
        self, func: Union[CheckReal, CheckFloat], x: float
    ) -> None:
        assert func(x)
        assert func(x, consider=None)  # default
        assert func(x, consider=fastnumbers.NUMBER_ONLY)

    funcs = identification_funcs

    @given(integers() | floats())
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_false_if_given_number_and_str_only_is_true(
        self, func: IdentificationFuncs, x: FloatOrInt
    ) -> None:
        assert not func(x, consider=fastnumbers.STRING_ONLY)

    # Handling of strings containing numbers as input

    funcs = ["check_real", "check_float"]

    @given(floats(allow_nan=False, allow_infinity=False).map(repr))
    @example("10." + "0" * 1050)
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_true_if_given_float_string(
        self, func: Union[CheckReal, CheckFloat], x: str
    ) -> None:
        assert func(x)
        assert func(pad(x))  # Accepts padding
        assert func(x, consider=None)  # default
        assert func(x, consider=fastnumbers.STRING_ONLY)

    funcs = identification_funcs

    @given(integers().map(repr))
    @example("40992764608243448035")
    @example("-41538374848935286698640072416676709")
    @example("240278958776173358420034462324117625982")
    @example("1609422692302207451978552816956662956486")
    @example("-121799354242674784350540853922878239740762834")
    @example("32718704454132572934419741118153895444518280065843028297496525078")
    @example("33684944745210074227862907273261282807602986571245071790093633147269")
    @example("1" + "0" * 1050)
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_true_if_given_int_string(
        self, func: IdentificationFuncs, x: str
    ) -> None:
        assert func(x)
        assert func(pad(x))  # Accepts padding
        assert func(x, consider=None)  # default
        assert func(x, consider=fastnumbers.STRING_ONLY)

    @given((integers() | floats(allow_nan=False, allow_infinity=False)).map(repr))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_false_if_given_string_and_num_only_is_true(
        self, func: IdentificationFuncs, x: str
    ) -> None:
        assert not func(x, consider=fastnumbers.NUMBER_ONLY)

    @given(sampled_from(digits))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_digit_returns_true(
        self, func: IdentificationFuncs, x: str
    ) -> None:
        assert func(x)
        assert func(pad(x))  # Accepts padding

    funcs = ["check_real", "check_float"]

    @given(sampled_from(numeric_not_digit_not_int))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_numeral_returns_true(
        self, func: Union[CheckReal, CheckFloat], x: str
    ) -> None:
        assert func(x)
        assert func(pad(x))  # Accepts padding

    # Handling of invalid input

    funcs = identification_funcs

    @given((text() | binary()).filter(not_a_number))
    @example("+")
    @example("-")
    @example("e")
    @example("e8")
    @example(".")
    @example(b"0e\x00")
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_false_if_given_non_number_string(
        self, func: IdentificationFuncs, x: str
    ) -> None:
        assert not func(x)

    @given(sampled_from(random.sample(not_numeric, 1000)))
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_non_numeral_returns_false(
        self, func: IdentificationFuncs, x: str
    ) -> None:
        assert not func(x)

    @given(text(min_size=2).filter(not_a_number))
    @example("   \u2007\u2007    ")
    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_given_unicode_of_more_than_one_char_returns_false(
        self, func: IdentificationFuncs, x: str
    ) -> None:
        assert not func(x)

    funcs = ["check_int", "check_intlike"]

    @parametrize("func", get_funcs(funcs), ids=funcs)
    def test_returns_false_for_nan_or_inf_string(
        self, func: Union[CheckInt, CheckIntLike]
    ) -> None:
        assert not func("nan")
        assert not func("inf")


class TestCheckFloat:
    """
    Tests for the check_float function that are too specific for the generalized tests.
    """

    @given(integers())
    def test_returns_false_if_given_int(self, x: int) -> None:
        assert not fastnumbers.check_float(x)

    @given(integers().map(repr))
    def test_returns_false_if_given_int_string_and_strict_true(self, x: str) -> None:
        assert not fastnumbers.check_float(x, strict=True)

    @given(integers().map(repr))
    def test_returns_true_if_given_int_string_and_strict_false(self, x: str) -> None:
        assert fastnumbers.check_float(x)
        assert fastnumbers.check_float(x, strict=False)  # default


class TestCheckInt:
    """
    Tests for the check_int function that are too specific for the generalized tests.
    """

    @given(floats())
    def test_returns_false_if_given_float(self, x: str) -> None:
        assert not fastnumbers.check_int(x)

    @given(integers())
    def test_returns_true_if_given_int_with_base(self, x: int) -> None:
        for base in range(2, 36 + 1):
            assert fastnumbers.check_int(x, base=base)

    @given(integers())
    def test_returns_true_if_given_int_string_with_non_base_10(self, x: int) -> None:
        for base in range(2, 36 + 1):
            # Avoid recursion error because of overly simple baseN function.
            # Avoid cases where number ends up creating infinity
            actual = base_n(x, base)
            if len(repr(x)) < 30 and isinstance(actual, int):
                assert fastnumbers.check_int(base_n(x, base), base=base)
        assert fastnumbers.check_int(bin(x), base=2)
        assert fastnumbers.check_int(bin(x), base=0)
        assert fastnumbers.check_int(oct(x), base=8)
        assert fastnumbers.check_int(oct(x), base=0)
        assert fastnumbers.check_int(oct(x).replace("0o", "0"), base=8)
        if x != 0:
            assert not fastnumbers.check_int(oct(x).replace("0o", "0"), base=0)
        assert fastnumbers.check_int(hex(x), base=16)
        assert fastnumbers.check_int(hex(x), base=0)
        # Force unicode path
        assert fastnumbers.check_int(hex(x).replace("0", "\uFF10"), base=0)

    def test_underscores(self) -> None:
        assert fastnumbers.check_int("0_0_0", allow_underscores=True)
        assert fastnumbers.check_int("0_0_0", base=0, allow_underscores=True)
        assert fastnumbers.check_int("4_2", allow_underscores=True)
        assert fastnumbers.check_int("4_2", base=0, allow_underscores=True)
        assert fastnumbers.check_int("1_0000_0000", allow_underscores=True)
        assert fastnumbers.check_int("1_0000_0000", base=0, allow_underscores=True)
        assert fastnumbers.check_int("0b1001_0100", base=0, allow_underscores=True)
        assert fastnumbers.check_int("0xffff_ffff", base=0, allow_underscores=True)
        assert fastnumbers.check_int("0o5_7_7", base=0, allow_underscores=True)
        assert fastnumbers.check_int("0b_0", base=0, allow_underscores=True)
        assert fastnumbers.check_int("0x_f", base=0, allow_underscores=True)
        assert fastnumbers.check_int("0o_5", base=0, allow_underscores=True)

        # Underscores in the base selector:
        assert not fastnumbers.check_int("0_b0", allow_underscores=True)
        assert not fastnumbers.check_int("0_b0", base=0, allow_underscores=True)
        assert not fastnumbers.check_int("0_xf", allow_underscores=True)
        assert not fastnumbers.check_int("0_xf", base=0, allow_underscores=True)
        assert not fastnumbers.check_int("0_o5", allow_underscores=True)
        assert not fastnumbers.check_int("0_o5", base=0, allow_underscores=True)

        # Old-style octal, still disallowed if base guess is needed:
        assert not fastnumbers.check_int("0_7", base=0, allow_underscores=True)
        assert not fastnumbers.check_int("09_99", base=0, allow_underscores=True)

        # Two underscores:
        assert not fastnumbers.check_int("0b1001__0100", base=0, allow_underscores=True)
        assert not fastnumbers.check_int("0xffff__ffff", base=0, allow_underscores=True)

    @given(floats(allow_nan=False, allow_infinity=False).map(repr))
    def test_returns_false_if_given_float_string(self, x: str) -> None:
        assert not fastnumbers.check_int(x)
        assert not fastnumbers.check_int(pad(x))
        for base in range(2, 36 + 1):
            if len(x) < 30:
                assert not fastnumbers.check_int(x, base=base)

    @given(sampled_from(numeric_not_digit_not_int))
    def test_given_unicode_numeral_returns_false(self, x: str) -> None:
        assert not fastnumbers.check_int(x)


class TestCheckIntLike:
    """
    Tests for the check_intlike function that are too specific for the generalized tests.
    """

    @given(floats().filter(not_an_integer))
    def test_returns_false_if_given_non_integer_float(self, x: float) -> None:
        assert not fastnumbers.check_intlike(x)

    @given(floats().filter(an_integer))
    def test_returns_true_if_given_integer_float(self, x: float) -> None:
        assert fastnumbers.check_intlike(x)

    @given(
        floats(allow_nan=False, allow_infinity=False).filter(not_an_integer).map(repr)
    )
    def test_returns_false_if_given_non_integer_float_string(self, x: str) -> None:
        assert not fastnumbers.check_intlike(x)
        assert not fastnumbers.check_intlike(pad(x))  # Accepts padding

    @given(sampled_from(numeric_not_digit_not_int))
    def test_given_unicode_non_digit_numeral_returns_false(self, x: str) -> None:
        assert not fastnumbers.check_intlike(x)

    @given(
        sampled_from(numeric_not_digit).filter(
            lambda x: an_integer(unicodedata.numeric(x))
        )
    )
    def test_given_unicode_digit_numeral_returns_true(self, x: str) -> None:
        assert fastnumbers.check_intlike(x)
        assert fastnumbers.check_intlike(pad(x))  # Accepts padding


class TestQueryType:
    """Tests for the query_type function."""

    def test_allowed_type_must_be_a_sequence(self) -> None:
        with raises(TypeError, match="allowed_type is not a sequence type"):
            fastnumbers.query_type("5", allowed_types={str: float})  # type: ignore

    def test_allowed_type_must_non_empty(self) -> None:
        with raises(ValueError, match="allowed_type must not be an empty sequence"):
            fastnumbers.query_type("5", allowed_types=[])

    @given(integers())
    def test_returns_int_if_given_int(self, x: int) -> None:
        assert fastnumbers.query_type(x) is int

    @given(floats())
    def test_returns_float_if_given_float(self, x: float) -> None:
        assert fastnumbers.query_type(x) is float

    @given(integers())
    def test_returns_none_if_given_int_and_int_is_not_allowed(self, x: int) -> None:
        assert fastnumbers.query_type(x, allowed_types=(float,)) is None

    @given(floats())
    def test_returns_none_if_given_float_and_float_is_not_allowed(
        self, x: float
    ) -> None:
        assert fastnumbers.query_type(x, allowed_types=(int,)) is None

    @given(integers().map(repr))
    def test_returns_int_if_given_int_string(self, x: str) -> None:
        assert fastnumbers.query_type(x) is int
        assert fastnumbers.query_type(pad(x)) is int  # Accepts padding

    @given(floats(allow_nan=False, allow_infinity=False).map(repr))
    def test_returns_float_if_given_float_string_padded_or_not(self, x: str) -> None:
        assert fastnumbers.query_type(x) is float
        assert fastnumbers.query_type(pad(x)) is float  # Accpets padding

    @given(integers().map(repr))
    def test_returns_none_if_given_int_string_and_int_is_not_allowed(
        self, x: str
    ) -> None:
        assert fastnumbers.query_type(x, allowed_types=(float, str)) is None

    @given(sampled_from(digits))
    def test_given_unicode_digit_returns_int(self, x: str) -> None:
        assert fastnumbers.query_type(x) is int
        assert fastnumbers.query_type(pad(x)) is int  # Accepts padding

    @given(sampled_from(numeric_not_digit_not_int))
    def test_given_unicode_numeral_returns_float(self, x: str) -> None:
        assert fastnumbers.query_type(x) is float
        assert fastnumbers.query_type(pad(x)) is float  # Accepts padding

    @given(sampled_from(random.sample(not_numeric, 1000)))
    def test_given_unicode_non_numeral_returns_str_or_none_if_str_not_allowed(
        self, x: str
    ) -> None:
        assert fastnumbers.query_type(x) is str
        assert fastnumbers.query_type(x, allowed_types=(int, float)) is None

    @given(text(min_size=2).filter(not_a_number))
    @example("   \u2007\u2007    ")
    def test_given_unicode_of_more_than_one_char_returns_str(self, x: str) -> None:
        assert fastnumbers.query_type(x) is str

    @given(text().filter(not_a_number))
    @example("+")
    @example("-")
    @example("e")
    @example("e8")
    @example(".")
    def test_returns_str_if_given_non_number_string(self, x: str) -> None:
        assert fastnumbers.query_type(x) is str

    @given(text().filter(not_a_number))
    def test_returns_none_if_given_non_number_string_and_str_is_not_allowed(
        self, x: str
    ) -> None:
        assert fastnumbers.query_type(x, allowed_types=(int, float)) is None

    @given(binary().filter(not_a_number))
    def test_returns_bytes_if_given_non_number_string(self, x: bytes) -> None:
        assert fastnumbers.query_type(x) is bytes

    @given(binary().filter(not_a_number))
    def test_returns_none_if_given_non_number_bytes_and_bytes_is_not_allowed(
        self, x: bytes
    ) -> None:
        assert fastnumbers.query_type(x, allowed_types=(int, float)) is None

    @parametrize("x", all_nan + [pad("+nan"), pad("-NAN")])
    def test_returns_str_for_nan_string_unless_allow_nan_is_true(self, x: str) -> None:
        assert fastnumbers.query_type(x) is str
        assert fastnumbers.query_type(x, allow_nan=True) is float

    @parametrize("x", most_inf + neg_inf + [pad("+inf"), pad("-INFINITY")])
    def test_returns_str_for_inf_string_unless_allow_infinity_is_true(
        self, x: str
    ) -> None:
        assert fastnumbers.query_type(x) is str
        assert fastnumbers.query_type(x, allow_inf=True) is float

    def test_given_nan_returns_float(self) -> None:
        assert fastnumbers.query_type(float("nan")) is float

    def test_given_inf_returns_float(self) -> None:
        assert fastnumbers.query_type(float("inf")) is float

    @given(floats(allow_nan=False).filter(an_integer))
    def test_given_float_returns_int_if_intlike_with_coerce(self, x: float) -> None:
        assert fastnumbers.query_type(x, coerce=True) == int

    @given(floats(allow_nan=False))
    def test_given_float_returns_float_or_int_with_coerce(self, x: float) -> None:
        assert (
            fastnumbers.query_type(x, coerce=True) == int if x.is_integer() else float
        )

    @given(
        lists(floats())
        | tuples(floats())
        | dictionaries(floats(), floats())
        | sets(floats())
        | iterables(floats())
    )
    def test_containers_returns_container_type(
        self, x: Union[Dict[float, float], Iterable[float]]
    ) -> None:
        assert fastnumbers.query_type(x) is type(x)
        assert fastnumbers.query_type(x, allowed_types=(float, int, str)) is None


class TestMappingFunctions:
    """Ensure that mapping functions operate on iterables"""

    @given(lists(floats() | integers() | text(max_size=50), min_size=1, max_size=50))
    @parametrize(
        "nomapper, mapper",
        [
            (fastnumbers.try_real, partial(fastnumbers.try_real, map=list)),
            (fastnumbers.try_float, partial(fastnumbers.try_float, map=list)),
            (fastnumbers.try_int, partial(fastnumbers.try_int, map=list)),
            (fastnumbers.try_forceint, partial(fastnumbers.try_forceint, map=list)),
        ],
    )
    @parametrize(
        "kwargs",
        [
            {},
            {"inf": fastnumbers.RAISE},
            {"nan": fastnumbers.RAISE},
            {"on_fail": fastnumbers.RAISE},
            {"base": 16},
            {"coerce": True},
        ],
    )
    def test_mapping_non_mapping_behave_the_same(
        self,
        nomapper: ConversionFuncs,
        mapper: ConversionFuncs,
        kwargs: Dict[str, Any],
        x: List[Union[float, int, str]],
    ) -> None:
        nomapper = partial(nomapper, **kwargs)
        mapper = partial(mapper, **kwargs)
        nonmapping = capture_result(lambda y: list(map(nomapper, y)), x)  # type: ignore
        mapping = capture_result(mapper, x)
        if nonmapping != nonmapping and mapping != mapping:
            assert math.isnan(nonmapping) and math.isnan(mapping)
        else:
            assert nonmapping == mapping

    @parametrize(
        "nomapper, mapper",
        [
            (fastnumbers.try_real, partial(fastnumbers.try_real, map=list)),
            (fastnumbers.try_float, partial(fastnumbers.try_float, map=list)),
            (fastnumbers.try_int, partial(fastnumbers.try_int, map=list)),
            (fastnumbers.try_forceint, partial(fastnumbers.try_forceint, map=list)),
        ],
    )
    def test_mapping_non_mapping_behave_the_same_with_invalid_types(
        self, nomapper: ConversionFuncs, mapper: ConversionFuncs
    ) -> None:
        x = ["7", "5", None]
        nonmapping = capture_result(lambda y: list(map(nomapper, y)), x)  # type: ignore
        mapping = capture_result(mapper, x)
        assert nonmapping == mapping

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=list),
            partial(fastnumbers.try_float, map=list),
            partial(fastnumbers.try_int, map=list),
            partial(fastnumbers.try_forceint, map=list),
            partial(fastnumbers.try_real, map=True),
            partial(fastnumbers.try_float, map=True),
            partial(fastnumbers.try_int, map=True),
            partial(fastnumbers.try_forceint, map=True),
        ],
    )
    @parametrize(
        "iterable_gen",
        [
            lambda: ["6", "4", "590"],
            lambda: ("6", "4", "590"),
            lambda: {"6", "4", "590"},
            lambda: iter(["6", "4", "590"]),
            lambda: (x for x in ["6", "4", "590"]),
        ],
    )
    def test_mapping_handles_any_iterable(
        self, func: ConversionFuncs, iterable_gen: Callable[[], Iterable[Any]]
    ) -> None:
        expected = [4, 6, 590]
        result = sorted(func(iterable_gen()))  # sorted needed b/c of set
        assert result == expected

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=list),
            partial(fastnumbers.try_float, map=list),
            partial(fastnumbers.try_int, map=list),
            partial(fastnumbers.try_forceint, map=list),
        ],
    )
    def test_mapping_handles_range(self, func: ConversionFuncs) -> None:
        """Range is a sequence but is not a 'fast sequence'"""
        expected = [0, 1, 2, 3]
        result = func(range(4))
        assert result == expected

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=True),
            partial(fastnumbers.try_float, map=True),
            partial(fastnumbers.try_int, map=True),
            partial(fastnumbers.try_forceint, map=True),
        ],
    )
    def test_mapping_iterator_handles_range(self, func: ConversionFuncs) -> None:
        """Range is a sequence but is not a 'fast sequence'"""
        result = func(range(4))
        print(1)
        assert next(result) == 0
        print(2)
        assert next(result) == 1
        print(3)
        assert next(result) == 2
        print(4)
        assert next(result) == 3
        print(5)
        assert next(result, None) is None
        print(6)

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=list),
            partial(fastnumbers.try_float, map=list),
            partial(fastnumbers.try_int, map=list),
            partial(fastnumbers.try_forceint, map=list),
            partial(fastnumbers.try_real, map=True),
            partial(fastnumbers.try_float, map=True),
            partial(fastnumbers.try_int, map=True),
            partial(fastnumbers.try_forceint, map=True),
        ],
    )
    def test_mapping_handles_broken_generator(self, func: ConversionFuncs) -> None:
        """A generator's exception should be returned"""

        def broken() -> Iterator[str]:
            """Not a good generator"""
            yield "5"
            yield "6"
            raise ValueError("Fëanor")

        with raises(ValueError, match="Fëanor"):
            for _ in func(broken()):
                pass

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=list),
            partial(fastnumbers.try_float, map=list),
            partial(fastnumbers.try_int, map=list),
            partial(fastnumbers.try_forceint, map=list),
        ],
    )
    @parametrize(
        "iterable_gen",
        [
            lambda: [],
            lambda: (),
            lambda: set(),
            lambda: iter([]),
            lambda: (x for x in []),  # type: ignore [var-annotated]
        ],
    )
    def test_mapping_handles_empty_iterable(
        self, func: ConversionFuncs, iterable_gen: Callable[[], Iterable[Any]]
    ) -> None:
        expected: List[Any] = []
        result = func(iterable_gen())
        assert result == expected

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=True),
            partial(fastnumbers.try_float, map=True),
            partial(fastnumbers.try_int, map=True),
            partial(fastnumbers.try_forceint, map=True),
        ],
    )
    @parametrize(
        "iterable_gen",
        [
            lambda: [],
            lambda: (),
            lambda: set(),
            lambda: iter([]),
            lambda: (x for x in []),  # type: ignore [var-annotated]
        ],
    )
    def test_mapping_iterable_handles_empty_iterable(
        self, func: ConversionFuncs, iterable_gen: Callable[[], Iterable[Any]]
    ) -> None:
        with raises(StopIteration):
            next(func(iterable_gen()))

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=list),
            partial(fastnumbers.try_float, map=list),
            partial(fastnumbers.try_int, map=list),
            partial(fastnumbers.try_forceint, map=list),
            partial(fastnumbers.try_real, map=True),
            partial(fastnumbers.try_float, map=True),
            partial(fastnumbers.try_int, map=True),
            partial(fastnumbers.try_forceint, map=True),
        ],
    )
    def test_mapping_raises_type_error_on_non_iterable(
        self, func: ConversionFuncs
    ) -> None:
        with raises(TypeError, match="'int' object is not iterable"):
            func(5)

    @parametrize(
        "func",
        [
            partial(fastnumbers.try_real, map=list),
            partial(fastnumbers.try_float, map=list),
            partial(fastnumbers.try_int, map=list),
            partial(fastnumbers.try_forceint, map=list),
            partial(fastnumbers.try_real, map=True),
            partial(fastnumbers.try_float, map=True),
            partial(fastnumbers.try_int, map=True),
            partial(fastnumbers.try_forceint, map=True),
        ],
    )
    @parametrize("style", [list, iter])
    def test_invalid_types_behave_as_expected(
        self, func: ConversionFuncs, style: Callable[[Any], Any]
    ) -> None:
        with raises(TypeError, match="not 'tuple'"):
            for _ in func(style([("Fëanor",)])):
                pass
        expected = [5]
        result = list(func(style([("Fëanor",)]), on_type_error=5))
        assert result == expected



#! /usr/bin/env python

"""
Cross-platform clean of working directory.
INTENDED TO BE CALLED FROM PROJECT ROOT, NOT FROM dev/!
"""

import pathlib
import shutil

# Directories to obliterate
dirs = [
    pathlib.Path("build"),
    pathlib.Path("dist"),
    pathlib.Path(".pytest_cache"),
    pathlib.Path(".hypothesis"),
    pathlib.Path(".tox"),
]
dirs += pathlib.Path.cwd().glob("*.egg-info")
for d in dirs:
    if d.is_dir():
        shutil.rmtree(d, ignore_errors=True)
    elif d.is_file():
        d.unlink()  # just in case there is a file.

# Clean up any stray __pycache__.
for d in pathlib.Path.cwd().rglob("__pycache__"):
    shutil.rmtree(d, ignore_errors=True)

# Patched doctests should be removed
for f in pathlib.Path.cwd().glob("doctest.py*"):
    f.unlink()

# Shouldn't be any .pyc left, but just in case
for f in pathlib.Path.cwd().rglob("*.pyc"):
    f.unlink()



#! /usr/bin/env python

"""
Cross-platform checking if code is appropriately formatted.
INTENDED TO BE CALLED FROM PROJECT ROOT, NOT FROM dev/!
"""

import glob
import shlex
import subprocess
import sys

check = "--check" in sys.argv

# Check that python code is formatted
black = ["black"]
if check:
    black.extend(["--quiet", "--check", "--diff"])
black.append(".")
print(*map(shlex.quote, black))
black_ret = subprocess.run(black)

# Check that C++ code is formatted
clang_format = [
    "clang-format",
    "--style=file:dev/clang-format.cfg",
    "--dry-run" if check else "-i",
]
cpp = glob.glob("src/cpp/*.cpp")
hpp = glob.glob("include/fastnumbers/*.hpp")
hpp += glob.glob("include/fastnumbers/parser/*.hpp")

clang_format = clang_format + cpp + hpp
print(*map(shlex.quote, clang_format))
clang_format_ret = subprocess.run(
    clang_format,
    stdout=subprocess.PIPE,
    stderr=subprocess.STDOUT,
    universal_newlines=True,
)
print(clang_format_ret.stdout)

# Stop here if not checking
if not check:
    sys.exit(0)

# Did any C++ files get formatted?
any_cpp_formatting = any(
    "clang-format-violations" in line for line in clang_format_ret.stdout.splitlines()
)

# "Roll up" what happened into a single exit code.
all_return_zero = all(ret.returncode == 0 for ret in [black_ret, clang_format_ret])
if any_cpp_formatting or not all_return_zero:
    sys.exit("Not all files are formatted correctly. Run 'tox -e format'.")
else:
    sys.exit(0)



#! /usr/bin/env python
"""
This script will run tests using gdb as a means to catch segfaults.

If gdb is not installed, it just runs the tests.
"""

import os
import sys

my_python = sys.argv[1]
other_args = sys.argv[2:]
try:
    # Don't use gdb unless requesting debugging mode
    if "FN_DEBUG" not in os.environ:
        raise OSError

    # Attempt to run pytest with debugger
    os.execlp(
        "gdb",
        "gdb",
        "-ex",
        "run",
        "-ex",
        "bt",
        "-ex",
        "quit",
        "--args",
        my_python,
        "-m",
        "pytest",
        "--doctest-glob=README.rst",
        *other_args,
    )
except OSError:
    # No debugger installed, just run pytest directly
    os.execl(
        my_python, my_python, "-m", "pytest", "--doctest-glob=README.rst", *other_args
    )



from __future__ import annotations

import array
import ctypes
from typing import Any, Callable, Dict, Iterator, List, NoReturn, Tuple, Union

import numpy as np
import pytest
from hypothesis import given as hyp_given
from hypothesis.strategies import (
    binary,
    floats,
    integers,
    lists,
    text,
)
from typing_extensions import TypedDict

import fastnumbers
from conftest import base_n


# Map supported data types to the Python array internal format designator
formats = {
    "signed char": "b",
    "unsigned char": "B",
    "signed short": "h",
    "unsigned short": "H",
    "signed int": "i",
    "unsigned int": "I",
    "signed long": "l",
    "unsigned long": "L",
    "signed long long": "q",
    "unsigned long long": "Q",
    "float": "f",
    "double": "d",
}

# Extreme values for signed integers of various byte widths
signed_size_extreme = {
    1: (-128, 127),
    2: (-32768, 32767),
    4: (-2147483648, 2147483647),
    8: (-9223372036854775808, 9223372036854775807),
}

# Extreme values for unsigned integers of various byte widths
unsigned_size_extreme = {
    1: (0, 255),
    2: (0, 65535),
    4: (0, 4294967295),
    8: (0, 18446744073709551615),
}

# Map all supported data types to the extreme values they support
extremes = {
    "signed char": signed_size_extreme[ctypes.sizeof(ctypes.c_byte)],
    "unsigned char": unsigned_size_extreme[ctypes.sizeof(ctypes.c_ubyte)],
    "signed short": signed_size_extreme[ctypes.sizeof(ctypes.c_short)],
    "unsigned short": unsigned_size_extreme[ctypes.sizeof(ctypes.c_ushort)],
    "signed int": signed_size_extreme[ctypes.sizeof(ctypes.c_int)],
    "unsigned int": unsigned_size_extreme[ctypes.sizeof(ctypes.c_uint)],
    "signed long": signed_size_extreme[ctypes.sizeof(ctypes.c_long)],
    "unsigned long": unsigned_size_extreme[ctypes.sizeof(ctypes.c_ulong)],
    "signed long long": signed_size_extreme[ctypes.sizeof(ctypes.c_longlong)],
    "unsigned long long": unsigned_size_extreme[ctypes.sizeof(ctypes.c_ulonglong)],
}

float_extremes = {
    "float": (1.17549e-38, 3.40282e38),
    "double": (2.22507e-308, 1.79769e308),
}

# Create shortcuts to collections of data types to test
signed_data_types = [
    "signed char",
    "signed short",
    "signed int",
    "signed long",
    "signed long long",
]
unsigned_data_types = [
    "unsigned char",
    "unsigned short",
    "unsigned int",
    "unsigned long",
    "unsigned long long",
]
int_data_types = signed_data_types + unsigned_data_types
float_data_types = ["float", "double"]
data_types = int_data_types + float_data_types


def test_invalid_argument_raises_type_error() -> None:
    given = [0, 1]
    with pytest.raises(TypeError, match="got an unexpected keyword argument 'invalid'"):
        fastnumbers.try_array(given, invalid="dummy")  # type: ignore


@pytest.mark.parametrize(
    "selector",
    [
        fastnumbers.RAISE,
        fastnumbers.INPUT,
        fastnumbers.DISALLOWED,
        fastnumbers.NUMBER_ONLY,
        fastnumbers.STRING_ONLY,
    ],
)
def test_selectors_are_rejected_when_invalid_for_inf_and_nan(selector: object) -> None:
    with pytest.raises(ValueError, match="values for 'inf' and 'nan'"):
        fastnumbers.try_array(["5"], inf=selector)
    with pytest.raises(ValueError, match="values for 'inf' and 'nan'"):
        fastnumbers.try_array(["5"], nan=selector)


@pytest.mark.parametrize(
    "selector",
    [
        fastnumbers.ALLOWED,
        fastnumbers.INPUT,
        fastnumbers.DISALLOWED,
        fastnumbers.NUMBER_ONLY,
        fastnumbers.STRING_ONLY,
    ],
)
def test_selectors_are_rejected_when_invalid_for_on_fail_and_friends(
    selector: object,
) -> None:
    msg = "values for 'on_fail', 'on_overflow', and 'on_type_error'"
    with pytest.raises(ValueError, match=msg):
        fastnumbers.try_array(["5"], on_fail=selector)
    with pytest.raises(ValueError, match=msg):
        fastnumbers.try_array(["5"], on_overflow=selector)
    with pytest.raises(ValueError, match=msg):
        fastnumbers.try_array(["5"], on_type_error=selector)


def test_invalid_input_type_gives_type_error() -> None:
    """Giving an invalid output type is rejected"""
    given = [0, 1]
    expected = "Only numpy ndarray and array.array types for output are "
    expected += r"supported, not <class 'list'>"
    with pytest.raises(TypeError, match=expected):
        fastnumbers.try_array(given, [])  # type: ignore


def test_require_output_if_numpy_is_not_installed() -> None:
    """A missing output value requires numpy to construct the output"""
    given = [0, 1]
    orig = fastnumbers.has_numpy
    try:
        fastnumbers.has_numpy = False
        with pytest.raises(RuntimeError, match="To use fastnumbers"):
            fastnumbers.try_array(given)
    finally:
        fastnumbers.has_numpy = orig


@pytest.mark.parametrize(
    "output",
    [np.array([[0, 0], [0, 0]]), np.array(0)],
)
def test_require_identically_one_ndarray_dimension(
    output: np.ndarray[Any, np.dtype[np.int_]]
) -> None:
    with pytest.raises(ValueError, match="Can only accept arrays of dimension 1"):
        fastnumbers.try_array([0, 9], output)


def test_require_input_and_output_to_have_equal_size() -> None:
    output = array.array("d", [0, 0, 0])
    with pytest.raises(ValueError, match="input/output must be of equal size"):
        fastnumbers.try_array([0, 9], output)


# Not all dtypes exist on all platforms. Add only the ones to test
# here that exist.
other_dtypes = [
    getattr(np, x)
    for x in ("float128", "complex128", "half", "bool_", "bytes_", "str_")
    if hasattr(np, x)
]


@pytest.mark.parametrize("dtype", other_dtypes)
def test_invalid_numpy_dtypes_raises_correct_type_error(
    dtype: np.dtype[Any],
) -> None:
    """Numpy arrays as output with invalid dtypes give an error"""
    given = [0, 1]
    output = np.array([0, 0], dtype=dtype)
    expected = "The only supported numpy dtypes for output are: "
    with pytest.raises(TypeError, match=expected):
        fastnumbers.try_array(given, output)


class TestCPPProtections:
    """
    These tests check internal C++ error handling that should not
    be possible to encounter from the python interface
    """

    def test_non_memorybuffer_type_raises_correct_type_error(self) -> None:
        """Ensure we only accept well-behaved memory views as input"""
        with pytest.raises(TypeError, match="not 'list'"):
            fastnumbers._array([0, 1], [0, 0])  # type: ignore

    @pytest.mark.parametrize("dtype", other_dtypes)
    def test_invalid_memorybuffer_type_raises_correct_type_error(
        self, dtype: np.dtype[Any]
    ) -> None:
        """Ensure we only accept well-behaved memory views as input"""
        given = [0, 1]
        output = np.array([0, 0], dtype=dtype)
        exception = r"Unknown buffer format '\S+' for object"
        with pytest.raises(TypeError, match=exception):
            fastnumbers._array(given, output)  # type: ignore


kwargs = ["inf", "nan", "on_fail", "on_overflow", "on_type_error"]
KwargsType = TypedDict(
    "KwargsType",
    {"inf": int, "nan": int, "on_fail": int, "on_overflow": int, "on_type_error": int},
    total=False,
)


class TestReplacements:
    """Test that the replacement mechanism works and raises intelligent exceptions"""

    @pytest.mark.parametrize("data_type", data_types)
    @pytest.mark.parametrize("kwarg", kwargs)
    def test_string_replacement_type_gives_type_error(
        self, data_type: str, kwarg: str
    ) -> None:
        given = [0, 1]
        result = array.array(formats[data_type], [0, 0])
        expected = f"The default value of 'not ok' given to option '{kwarg}' "
        expected += "has type 'str' which cannot be converted to a numeric value"
        with pytest.raises(TypeError, match=expected):
            fastnumbers.try_array(given, result, **{kwarg: "not ok"})  # type: ignore

    @pytest.mark.parametrize("data_type", int_data_types)
    @pytest.mark.parametrize("kwarg", kwargs)
    def test_float_replacement_type_for_int_gives_value_error(
        self, data_type: str, kwarg: str
    ) -> None:
        given = [0, 1]
        result = array.array(formats[data_type], [0, 0])
        expected = rf"The default value of 1\.3 given to option '{kwarg}' "
        expected += f"cannot be converted to C type '{data_type}'"
        with pytest.raises(ValueError, match=expected):
            fastnumbers.try_array(given, result, **{kwarg: 1.3})  # type: ignore

    @pytest.mark.parametrize("data_type", float_data_types)
    @pytest.mark.parametrize(
        "kwargs",
        [
            KwargsType(inf=0),
            KwargsType(nan=0),
            KwargsType(on_fail=0),
            KwargsType(on_overflow=0),
            KwargsType(on_type_error=0),
        ],
    )
    def test_int_replacement_type_for_float_is_ok(
        self, data_type: str, kwargs: KwargsType
    ) -> None:
        given = [0, 1]
        result = array.array(formats[data_type], [0, 0])
        expected = array.array(formats[data_type], [0.0, 1.0])
        fastnumbers.try_array(given, result, **kwargs)
        assert result == expected

    @pytest.mark.parametrize("data_type", int_data_types)
    @pytest.mark.parametrize("kwarg", kwargs)
    def test_extreme_replacement_type_for_int_gives_overflow_error(
        self, data_type: str, kwarg: str
    ) -> None:
        given = [0, 1]
        result = array.array(formats[data_type], [0, 0])
        expected = rf"The default value of \S+ given to option '{kwarg}' "
        expected += f"cannot be converted to C type '{data_type}' without overflowing"
        kwargs: KwargsType = {kwarg: extremes[data_type][0] - 1}  # type: ignore[misc]
        with pytest.raises(OverflowError, match=expected):
            fastnumbers.try_array(given, result, **kwargs)
        kwargs = {kwarg: extremes[data_type][1] + 1}  # type: ignore[misc]
        with pytest.raises(OverflowError, match=expected):
            fastnumbers.try_array(given, result, **kwargs)

    @pytest.mark.parametrize("data_type", data_types)
    def test_replacement_callables_with_invalid_type_gives_type_error(
        self, data_type: str
    ) -> None:
        given = ["invalid"]
        result = array.array(formats[data_type], [0])
        expected = "Callable passed to 'on_fail' with input 'invalid' returned "
        expected += "the value '3' that has type 'str' which cannot be converted "
        expected += "to a numeric value"
        with pytest.raises(TypeError, match=expected):
            fastnumbers.try_array(given, result, on_fail=lambda _: "3")

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_replacement_callables_with_bad_number_gives_value_error(
        self, data_type: str
    ) -> None:
        given = ["invalid"]
        result = array.array(formats[data_type], [0])
        expected = "Callable passed to 'on_fail' with input 'invalid' returned "
        expected += f"the value 3.4 that cannot be converted to C type '{data_type}'"
        with pytest.raises(ValueError, match=expected):
            fastnumbers.try_array(given, result, on_fail=lambda _: 3.4)

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_replacement_callables_with_extreme_gives_overflow_error(
        self, data_type: str
    ) -> None:
        given = ["invalid"]
        result = array.array(formats[data_type], [0])
        expected = "Callable passed to 'on_fail' with input 'invalid' returned "
        expected += r"the value \S+ that cannot be converted to C type "
        expected += f"'{data_type}' without overflowing"
        with pytest.raises(OverflowError, match=expected):
            fastnumbers.try_array(
                given, result, on_fail=lambda _: extremes[data_type][1] + 1
            )

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_replacement_callables_some_python_error(self, data_type: str) -> None:
        given = ["invalid"]
        result = array.array(formats[data_type], [0])
        expected = "bad operand type for abs()"
        with pytest.raises(TypeError, match=expected):
            fastnumbers.try_array(given, result, on_fail=lambda x: abs(x))

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_replacements_trigger_new_values_for_int(self, data_type: str) -> None:
        given = [
            "invalid",
            [2],
            extremes[data_type][0] - 1,
            extremes[data_type][1] + 1,
            "7",
        ]
        result = array.array(formats[data_type], [0, 0, 0, 0, 0])
        expected = array.array(formats[data_type], [2, 3, 1, 1, 7])
        fastnumbers.try_array(given, result, on_fail=2, on_overflow=1, on_type_error=3)
        assert result == expected

    @pytest.mark.parametrize("data_type", float_data_types)
    def test_replacements_trigger_new_values_for_float(self, data_type: str) -> None:
        given = ["invalid", [2], "inf", "nan", float("inf"), float("nan"), "7"]
        result = array.array(formats[data_type], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        expected = array.array(formats[data_type], [2.0, 3.0, 5.0, 4.0, 5.0, 4.0, 7.0])
        fastnumbers.try_array(given, result, inf=5, nan=4, on_fail=2, on_type_error=3)
        assert result == expected

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_replacement_callables_trigger_new_values_for_int(
        self, data_type: str
    ) -> None:
        given = [
            "invalid",
            [2],
            extremes[data_type][0] - 1,
            extremes[data_type][1] + 1,
            "7",
        ]
        result = array.array(formats[data_type], [0, 0, 0, 0, 0])
        expected = array.array(formats[data_type], [2, 3, 1, 1, 7])
        fastnumbers.try_array(
            given,
            result,
            on_fail=lambda _: 2,
            on_overflow=lambda _: 1,
            on_type_error=lambda _: 3,
        )
        assert result == expected

    @pytest.mark.parametrize("data_type", float_data_types)
    def test_replacement_callables_trigger_new_values_for_float(
        self, data_type: str
    ) -> None:
        given = ["invalid", [2], "inf", "nan", float("inf"), float("nan"), "7"]
        result = array.array(formats[data_type], [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0])
        expected = array.array(formats[data_type], [2.0, 3.0, 5.0, 4.0, 5.0, 4.0, 7.0])
        fastnumbers.try_array(
            given,
            result,
            inf=lambda _: 5,
            nan=lambda _: 4,
            on_fail=lambda _: 2,
            on_type_error=lambda _: 3,
        )
        assert result == expected


class DumbFloatClass(object):
    def __float__(self) -> NoReturn:
        raise ValueError("something here might go wrong")


class DumbIntClass(object):
    def __int__(self) -> NoReturn:
        raise ValueError("something here might go wrong")


class TestErrors:
    """Test that intelligent exceptions are raised on error"""

    @pytest.mark.parametrize("dumb", [DumbFloatClass(), DumbIntClass()])
    @pytest.mark.parametrize("data_type", data_types)
    @pytest.mark.parametrize("style", [list, iter])
    def test_given_junk_float_type_raises_error(
        self, data_type: str, dumb: Any, style: Callable[[Any], Any]
    ) -> None:
        given = style([dumb])
        result = array.array(formats[data_type], [0])
        expected = "Cannot convert"
        with pytest.raises(ValueError, match=expected):
            fastnumbers.try_array(given, result)

    @pytest.mark.parametrize("data_type", data_types)
    def test_given_invalid_string_raises_value_error(self, data_type: str) -> None:
        given = ["4", "78", "46", "invalid"]
        result = array.array(formats[data_type], [0, 0, 0, 0])
        expected = f"Cannot convert 'invalid' to C type '{data_type}'"
        with pytest.raises(ValueError, match=expected):
            fastnumbers.try_array(given, result)

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_given_extreme_raises_overflow_error(self, data_type: str) -> None:
        given = ["4", "78", "46", extremes[data_type][1] + 1]
        result = array.array(formats[data_type], [0, 0, 0, 0])
        expected = rf"Cannot convert \S+ to C type '{data_type}' without overflowing"
        with pytest.raises(OverflowError, match=expected):
            fastnumbers.try_array(given, result)

    @pytest.mark.parametrize("data_type", data_types)
    def test_given_incorrect_type_raises_type_error(self, data_type: str) -> None:
        given = ["4", "78", "46", ["6"]]
        result = array.array(formats[data_type], [0, 0, 0, 0])
        expected = r"The value \['6'\] has type 'list' which cannot "
        expected += "be converted to a numeric value"
        with pytest.raises(TypeError, match=expected):
            fastnumbers.try_array(given, result)

    @pytest.mark.parametrize("data_type", data_types)
    def test_given_broken_generator_fails(self, data_type: str) -> None:
        """A generator's exception should be returned"""

        def broken() -> Iterator[str]:
            """Not a good generator"""
            yield "5"
            yield "6"
            raise ValueError("Fëanor")

        output = array.array(formats[data_type], [0, 0, 0, 0])
        with pytest.raises(ValueError, match="Fëanor"):
            fastnumbers.try_array(broken(), output)

    @pytest.mark.parametrize("data_type", data_types)
    def test_given_non_iterable_raises_type_error(self, data_type: str) -> None:
        output = array.array(formats[data_type], [0, 0, 0, 0])
        with pytest.raises(TypeError, match="'int' object is not iterable"):
            fastnumbers.try_array(5, output)  # type: ignore

    @pytest.mark.parametrize("data_type", data_types)
    @pytest.mark.parametrize("style", [list, iter])
    def test_given_invalid_types_behave_as_expected(
        self, data_type: str, style: Callable[[Any], Any]
    ) -> None:
        msg = r"The value \('Fëanor',\) has type 'tuple' which cannot be "
        msg += "converted to a numeric value"
        with pytest.raises(TypeError, match=msg):
            fastnumbers.try_array(style([("Fëanor",)]))
        expected = array.array(formats[data_type], [5])
        result = array.array(formats[data_type], [0])
        fastnumbers.try_array(style([("Fëanor",)]), result, on_type_error=5)
        assert result == expected


class TestSuccess:
    """Test that the function does what it says on the tin"""

    @pytest.mark.parametrize("data_type", data_types)
    @pytest.mark.parametrize("style", [list, tuple, iter])
    def test_given_valid_values_returns_correct_results(
        self, data_type: str, style: Callable[[Any], Any]
    ) -> None:
        given = style(["4", "78", 46, "⑦"])
        result = array.array(formats[data_type], [0, 0, 0, 0])
        expected = array.array(formats[data_type], [4, 78, 46, 7])
        fastnumbers.try_array(given, result)
        assert result == expected

    @pytest.mark.parametrize("data_type", data_types)
    def test_given_range_returns_correct_results(self, data_type: str) -> None:
        given = range(4)
        result = array.array(formats[data_type], [0, 0, 0, 0])
        expected = array.array(formats[data_type], [0, 1, 2, 3])
        fastnumbers.try_array(given, result)
        assert result == expected

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_integer_extremes(self, data_type: str) -> None:
        given = [
            extremes[data_type][0],
            str(extremes[data_type][0]),
            extremes[data_type][0] - 1,
            str(extremes[data_type][0] - 1),
            extremes[data_type][1],
            str(extremes[data_type][1]),
            extremes[data_type][1] + 1,
            str(extremes[data_type][1] + 1),
        ]
        result = array.array(formats[data_type], [0, 0, 0, 0, 0, 0, 0, 0])
        expected = array.array(
            formats[data_type],
            [
                extremes[data_type][0],
                extremes[data_type][0],
                123,
                123,
                extremes[data_type][1],
                extremes[data_type][1],
                123,
                123,
            ],
        )
        fastnumbers.try_array(given, result, on_overflow=123)
        assert result == expected

    @pytest.mark.parametrize("data_type", float_data_types)
    def test_float_extremes(self, data_type: str) -> None:
        given = [
            float_extremes[data_type][0],
            str(float_extremes[data_type][0]),
            float_extremes[data_type][0] * 1e-100,  # so small, becomes 0.0
            str(float_extremes[data_type][0] * 1e-100),
            float_extremes[data_type][1],
            str(float_extremes[data_type][1]),
            float_extremes[data_type][1] * 1e10,  # so big, becomes infinity
            str(float_extremes[data_type][1] * 1e10),
        ]
        result = array.array(formats[data_type], [0, 0, 0, 0, 0, 0, 0, 0])
        expected = array.array(
            formats[data_type],
            [
                float_extremes[data_type][0],
                float_extremes[data_type][0],
                0.0,
                0.0,
                float_extremes[data_type][1],
                float_extremes[data_type][1],
                float("inf"),
                float("inf"),
            ],
        )
        fastnumbers.try_array(given, result)
        assert result == expected

    @pytest.mark.parametrize("base", range(2, 37))
    @pytest.mark.parametrize("data_type", int_data_types)
    def test_integer_bases(self, base: int, data_type: str) -> None:
        given = map(
            lambda x: base_n(x, base),
            [
                extremes[data_type][0],
                extremes[data_type][0] - 1,
                extremes[data_type][1],
                extremes[data_type][1] + 1,
                0,
                100,
            ],
        )
        result = array.array(formats[data_type], [0, 0, 0, 0, 0, 0])
        expected = array.array(
            formats[data_type],
            [
                extremes[data_type][0],
                123,
                extremes[data_type][1],
                123,
                0,
                100,
            ],
        )
        fastnumbers.try_array(given, result, base=base, on_overflow=123)
        assert result == expected

    @pytest.mark.parametrize("data_type", data_types)
    def test_underscores(self, data_type: str) -> None:
        given = ["1_0", "11_0"]
        result = array.array(formats[data_type], [0, 0])
        expected = array.array(formats[data_type], [10, 110])
        fastnumbers.try_array(given, result, allow_underscores=True)
        assert result == expected

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_base_prefix(self, data_type: str) -> None:
        given = [
            bin(extremes[data_type][0]),
            oct(extremes[data_type][0]),
            hex(extremes[data_type][0]),
        ]
        result = array.array(formats[data_type], [0, 0, 0])
        expected = array.array(formats[data_type], [extremes[data_type][0]] * 3)
        fastnumbers.try_array(given, result, base=0)
        assert result == expected

    @pytest.mark.parametrize("data_type", int_data_types)
    def test_base_prefix_and_underscores(self, data_type: str) -> None:
        given = ["0b10_01", "0x3_a", "0o5_7", "0b_0", "0x_f", "0o_5"]
        result = array.array(formats[data_type], [0, 0, 0, 0, 0, 0])
        expected = array.array(formats[data_type], [9, 58, 47, 0, 15, 5])
        fastnumbers.try_array(given, result, base=0, allow_underscores=True)
        assert result == expected


# Create shortcuts to collections of dtypes to test
signed_dtypes: List[Any] = [
    np.int8,
    np.int16,
    np.int32,
    np.int64,
]
unsigned_dtypes: List[Any] = [
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]
int_dtypes = signed_dtypes + unsigned_dtypes
float_dtypes: List[Any] = [np.float32, np.float64]
dtypes = int_dtypes + float_dtypes
dtype_extremes: Dict[Any, Tuple[int, int]] = {
    np.int8: signed_size_extreme[1],
    np.uint8: unsigned_size_extreme[1],
    np.int16: signed_size_extreme[2],
    np.uint16: unsigned_size_extreme[2],
    np.int32: signed_size_extreme[4],
    np.uint32: unsigned_size_extreme[4],
    np.int64: signed_size_extreme[8],
    np.uint64: unsigned_size_extreme[8],
}

dtype_float_extremes: Dict[Any, Tuple[float, float]] = {
    np.float32: (1.17549e-38, 3.40282e38),
    np.float64: (2.22507e-308, 1.79769e308),
}


class TestNumpy:
    """Ensure that try_array well supports numpy arrays"""

    def test_default_dtype_is_float64(self) -> None:
        given = [4, 4.5, "5", "5.6", "nan", "inf"]
        expected = np.array([4, 4.5, 5, 5.6, np.nan, np.inf], dtype=np.float64)
        result = fastnumbers.try_array(given)
        assert result.dtype == np.float64
        assert np.array_equal(result, expected, equal_nan=True)

    def test_accepts_iterable_as_input(self) -> None:
        given = iter([4, 4.5, "5", "5.6", "nan", "inf"])
        expected = np.array([4, 4.5, 5, 5.6, np.nan, np.inf], dtype=np.float64)
        result = fastnumbers.try_array(given)
        assert np.array_equal(result, expected, equal_nan=True)

    @pytest.mark.parametrize("dtype", dtypes)
    def test_supported_dtypes(
        self, dtype: Union[np.dtype[np.int_], np.dtype[np.float_]]
    ) -> None:
        given = [4, "5", "⑦"]
        expected = np.array([4, 5, 7], dtype=dtype)
        result = fastnumbers.try_array(given, dtype=dtype)
        assert result.dtype == dtype
        assert np.array_equal(result, expected)

    @pytest.mark.parametrize("dtype", int_dtypes)
    def test_integer_extremes(self, dtype: np.dtype[np.int_]) -> None:
        given = [
            dtype_extremes[dtype][0],
            str(dtype_extremes[dtype][0]),
            dtype_extremes[dtype][0] - 1,
            str(dtype_extremes[dtype][0] - 1),
            dtype_extremes[dtype][1],
            str(dtype_extremes[dtype][1]),
            dtype_extremes[dtype][1] + 1,
            str(dtype_extremes[dtype][1] + 1),
        ]
        expected = np.array(
            [
                dtype_extremes[dtype][0],
                dtype_extremes[dtype][0],
                123,
                123,
                dtype_extremes[dtype][1],
                dtype_extremes[dtype][1],
                123,
                123,
            ],
            dtype=dtype,
        )
        result = fastnumbers.try_array(given, dtype=dtype, on_overflow=123)
        assert np.array_equal(result, expected)

    @pytest.mark.parametrize("dtype", float_dtypes)
    def test_float_extremes(self, dtype: np.dtype[np.float_]) -> None:
        given = [
            dtype_float_extremes[dtype][0],
            str(dtype_float_extremes[dtype][0]),
            dtype_float_extremes[dtype][0] * 1e-100,  # so small, becomes 0.0
            str(dtype_float_extremes[dtype][0] * 1e-100),
            dtype_float_extremes[dtype][1],
            str(dtype_float_extremes[dtype][1]),
            dtype_float_extremes[dtype][1] * 1e10,  # so big, becomes infinity
            str(dtype_float_extremes[dtype][1] * 1e10),
        ]
        expected = np.array(
            [
                dtype_float_extremes[dtype][0],
                dtype_float_extremes[dtype][0],
                0.0,
                0.0,
                dtype_float_extremes[dtype][1],
                dtype_float_extremes[dtype][1],
                float("inf"),
                float("inf"),
            ],
            dtype=dtype,
        )
        result = fastnumbers.try_array(given, dtype=dtype)
        assert np.array_equal(result, expected)

    @pytest.mark.parametrize("dtype", dtypes)
    def test_accepts_output_array(
        self, dtype: Union[np.dtype[np.int_], np.dtype[np.float_]]
    ) -> None:
        given = [4, "5", "⑦"]
        result = np.array([0, 0, 0], dtype=dtype)
        expected = np.array([4, 5, 7], dtype=dtype)
        fastnumbers.try_array(given, result)
        assert np.array_equal(result, expected)

    def test_slice(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array([0, 0, 0, 0, 0])
        expected = np.array([0, 4, 5, 7, 0])
        fastnumbers.try_array(given, result[1:4])
        assert np.array_equal(result, expected)

    def test_strides(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array([0, 0, 0, 0, 0, 0])
        expected = np.array([4, 0, 5, 0, 7, 0])
        fastnumbers.try_array(given, result[::2])
        assert np.array_equal(result, expected)

    def test_strides_offset(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array([0, 0, 0, 0, 0, 0])
        expected = np.array([0, 4, 0, 5, 0, 7])
        fastnumbers.try_array(given, result[1::2])
        assert np.array_equal(result, expected)

    def test_negative_strides(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array([0, 0, 0, 0, 0, 0])
        expected = np.array([0, 7, 0, 5, 0, 4])
        fastnumbers.try_array(given, result[::-2])
        assert np.array_equal(result, expected)

    def test_slice_2d(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        expected = np.array([[0, 0, 0], [4, 5, 7], [0, 0, 0]])
        fastnumbers.try_array(given, result[1, :])
        assert np.array_equal(result, expected)

    def test_slice_2d_column(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
        expected = np.array([[0, 4, 0], [0, 5, 0], [0, 7, 0]])
        fastnumbers.try_array(given, result[:, 1])
        assert np.array_equal(result, expected)

    def test_stride_2d(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array([[0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0]])
        expected = np.array(
            [[0, 0, 0, 0, 0, 0], [4, 0, 5, 0, 7, 0], [0, 0, 0, 0, 0, 0]]
        )
        fastnumbers.try_array(given, result[1, ::2])
        assert np.array_equal(result, expected)

    def test_stride_2d_column(self) -> None:
        given = [4, "5", "⑦"]
        result = np.array(
            [[0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0], [0, 0, 0]]
        )
        expected = np.array(
            [[0, 0, 0], [0, 4, 0], [0, 0, 0], [0, 5, 0], [0, 0, 0], [0, 7, 0]]
        )
        fastnumbers.try_array(given, result[1::2, 1])
        assert np.array_equal(result, expected)


@hyp_given(
    lists(
        floats() | integers() | text() | binary() | lists(integers(), max_size=1),
        max_size=50,
    )
)
@pytest.mark.parametrize("dtype", int_dtypes)
def test_all_the_things_for_ints(dtype: np.dtype[np.int_], x: List[Any]) -> None:
    # Using try_array should give the same results
    # as try_int with map=list then converted to an array.
    # Under-the-hood, the on_fail, etc. replacements use a different code path
    # so this test is not just wasting time.
    expected_pre = fastnumbers.try_int(
        x,
        on_fail=lambda x: 5 if isinstance(x, str) else 6,
        on_type_error=7,
        map=list,
    )
    expected_pre = [9 if x < dtype_extremes[dtype][0] else x for x in expected_pre]
    expected_pre = [9 if x > dtype_extremes[dtype][1] else x for x in expected_pre]
    expected = np.array(expected_pre, dtype=dtype)
    result = fastnumbers.try_array(
        x,
        dtype=dtype,
        on_fail=lambda x: 5 if isinstance(x, str) else 6,
        on_type_error=7,
        on_overflow=9,
    )
    np.array_equal(result, expected)


@hyp_given(
    lists(
        floats() | integers() | text() | binary() | lists(integers(), max_size=1),
        max_size=50,
    )
)
@pytest.mark.parametrize("dtype", float_dtypes)
@pytest.mark.filterwarnings("ignore:overflow encountered in cast")
def test_all_the_things_for_floats(dtype: np.dtype[np.float_], x: List[Any]) -> None:
    # Using try_array should give the same results
    # as try_float with map=list then converted to an array.
    # Under-the-hood, the on_fail, etc. replacements use a different code path
    # so this test is not just wasting time.
    expected_pre = fastnumbers.try_float(
        x,
        on_fail=lambda x: 5.0 if isinstance(x, str) else 6.0,
        on_type_error=7.0,
        map=list,
    )
    expected = np.array(expected_pre, dtype=dtype)
    result = fastnumbers.try_array(
        x,
        dtype=dtype,
        on_fail=lambda x: 5.0 if isinstance(x, str) else 6.0,
        on_type_error=7.0,
    )
    np.array_equal(result, expected, equal_nan=True)


@hyp_given(
    lists(
        floats() | integers() | text() | binary() | lists(integers(), max_size=1),
        max_size=50,
    )
)
@pytest.mark.parametrize("dtype", float_dtypes)
@pytest.mark.filterwarnings("ignore:overflow encountered in cast")
def test_all_the_things_for_floats_with_nan_inf_replacement(
    dtype: np.dtype[np.float_], x: List[Any]
) -> None:
    # Using try_array should give the same results
    # as try_float with map=list then converted to an array.
    # Under-the-hood, the on_fail, etc. replacements use a different code path
    # so this test is not just wasting time.
    expected_pre = fastnumbers.try_float(
        x,
        inf=1.0,
        nan=3.0,
        on_fail=lambda x: 5.0 if isinstance(x, str) else 6.0,
        on_type_error=7.0,
        map=list,
    )
    expected = np.array(expected_pre, dtype=dtype)
    result = fastnumbers.try_array(
        x,
        dtype=dtype,
        inf=1.0,
        nan=3.0,
        on_fail=lambda x: 5.0 if isinstance(x, str) else 6.0,
        on_type_error=7.0,
    )
    np.array_equal(result, expected)


def test_sanity() -> None:
    """Make sure explictly passing dtypes makes mypy happy"""
    fastnumbers.try_array([0, 1], dtype=np.int8)
    fastnumbers.try_array([0, 1], dtype=np.int16)
    fastnumbers.try_array([0, 1], dtype=np.int32)
    fastnumbers.try_array([0, 1], dtype=np.int64)
    fastnumbers.try_array([0, 1], dtype=np.uint8)
    fastnumbers.try_array([0, 1], dtype=np.uint16)
    fastnumbers.try_array([0, 1], dtype=np.uint32)
    fastnumbers.try_array([0, 1], dtype=np.uint64)
    fastnumbers.try_array([0, 1], dtype=np.float32)
    fastnumbers.try_array([0, 1], dtype=np.float64)



import builtins
import sys
import time
import unittest
from math import copysign, isinf, isnan
from typing import Callable, List, Union

import builtin_support as support
from builtin_grammar import (
    INVALID_UNDERSCORE_LITERALS,
    VALID_UNDERSCORE_LITERALS,
)

from fastnumbers import float

INF = float("inf")
NAN = float("nan")

have_getformat = hasattr(float, "__getformat__")
requires_getformat = unittest.skipUnless(have_getformat, "requires __getformat__")
requires_setformat = unittest.skipUnless(
    hasattr(float, "__setformat__"), "requires __setformat__"
)


class FloatSubclass(builtins.float):
    pass


class OtherFloatSubclass(builtins.float):
    pass


class GeneralFloatCases(unittest.TestCase):
    def test_float(self) -> None:
        self.assertEqual(float(3.14), 3.14)
        self.assertEqual(float(314), 314.0)
        self.assertEqual(float("  3.14  "), 3.14)
        self.assertRaises(ValueError, float, "  0x3.1  ")
        self.assertRaises(ValueError, float, "  -0x3.p-1  ")
        self.assertRaises(ValueError, float, "  +0x3.p-1  ")
        self.assertRaises(ValueError, float, "++3.14")
        self.assertRaises(ValueError, float, "+-3.14")
        self.assertRaises(ValueError, float, "-+3.14")
        self.assertRaises(ValueError, float, "--3.14")
        self.assertRaises(ValueError, float, ".nan")
        self.assertRaises(ValueError, float, "+.inf")
        self.assertRaises(ValueError, float, ".")
        self.assertRaises(ValueError, float, "-.")
        self.assertRaises(TypeError, float, {})
        self.assertRaisesRegex(TypeError, "not 'dict'", float, {})
        # Lone surrogate
        if sys.version_info >= (3, 7):
            self.assertRaises(ValueError, float, "\uD8F0")
        # check that we don't accept alternate exponent markers
        self.assertRaises(ValueError, float, "-1.7d29")
        self.assertRaises(ValueError, float, "3D-14")
        self.assertEqual(float("  \u0663.\u0661\u0664  "), 3.14)
        self.assertEqual(float("\N{EM SPACE}3.14\N{EN SPACE}"), 3.14)
        # extra long strings should not be a problem
        float(b"." + b"1" * 1000)
        float("." + "1" * 1000)
        if sys.version_info >= (3, 7):
            # Invalid unicode string
            # See bpo-34087
            self.assertRaises(ValueError, float, "\u3053\u3093\u306b\u3061\u306f")

    def test_noargs(self) -> None:
        self.assertEqual(float(), 0.0)

    def test_underscores(self) -> None:
        for lit in VALID_UNDERSCORE_LITERALS:
            if not any(ch in lit for ch in "jJxXoObB"):
                self.assertEqual(float(lit), eval(lit))
                self.assertEqual(float(lit), float(lit.replace("_", "")))
        for lit in INVALID_UNDERSCORE_LITERALS:
            if lit in ("0_7", "09_99"):  # octals are not recognized here
                continue
            if not any(ch in lit for ch in "jJxXoObB"):
                self.assertRaises(ValueError, float, lit)
        # Additional test cases; nan and inf are never valid as literals,
        # only in the float() constructor, but we don't allow underscores
        # in or around them.
        self.assertRaises(ValueError, float, "_NaN")
        self.assertRaises(ValueError, float, "Na_N")
        self.assertRaises(ValueError, float, "IN_F")
        self.assertRaises(ValueError, float, "-_INF")
        self.assertRaises(ValueError, float, "-INF_")
        # Check that we handle bytes values correctly.
        self.assertRaises(ValueError, float, b"0_.\xff9")

    def test_non_numeric_input_types(self) -> None:
        # Test possible non-numeric types for the argument x, including
        # subclasses of the explicitly documented accepted types.
        class CustomStr(str):
            pass

        class CustomBytes(bytes):
            pass

        class CustomByteArray(bytearray):
            pass

        factories: List[Callable[[bytes], Union[bytes, bytearray, str]]] = [
            bytes,
            bytearray,
            lambda b: CustomStr(b.decode()),
            CustomBytes,
            CustomByteArray,
            memoryview,
        ]
        try:
            from array import array
        except ImportError:
            pass
        else:
            factories.append(lambda b: array("B", b))  # type: ignore

        for f in factories:
            x = f(b" 3.14  ")
            with self.subTest(type(x)):
                self.assertEqual(float(x), 3.14)
                with self.assertRaisesRegex(ValueError, "could not convert"):
                    float(f(b"A" * 0x10))

    def test_float_memoryview(self) -> None:
        self.assertEqual(float(memoryview(b"12.3")[1:4]), 2.3)
        self.assertEqual(float(memoryview(b"12.3\x00")[1:4]), 2.3)
        self.assertEqual(float(memoryview(b"12.3 ")[1:4]), 2.3)
        self.assertEqual(float(memoryview(b"12.3A")[1:4]), 2.3)
        self.assertEqual(float(memoryview(b"12.34")[1:4]), 2.3)

    def test_error_message_old(self) -> None:
        testlist = ("\xbd", "123\xbd", "  123 456  ")
        for s in testlist:
            try:
                float(s)
            except ValueError as e:
                self.assertIn(s.strip(), e.args[0])
            else:
                self.fail("Expected int(%r) to raise a ValueError" % s)

    def test_error_message(self) -> None:
        def check(s: Union[str, bytes]) -> None:
            with self.assertRaises(ValueError, msg="float(%r)" % (s,)) as cm:
                float(s)
            self.assertEqual(
                str(cm.exception), "could not convert string to float: %r" % (s,)
            )

        check("\xbd")
        check("123\xbd")
        check("  123 456  ")
        check(b"  123 456  ")
        # all whitespace (cf. https://github.com/python/cpython/issues/95605)
        check("")
        check(" ")
        check("\t \n")

        # non-ascii digits (error came from non-digit '!')
        check("\u0663\u0661\u0664!")
        # embedded NUL
        check("123\x00")
        check("123\x00 245")
        check("123\x00245")
        # byte string with embedded NUL
        check(b"123\x00")
        # non-UTF-8 byte string
        check(b"123\xa0")

    @support.run_with_locale("LC_NUMERIC", "fr_FR", "de_DE")
    def test_float_with_comma(self) -> None:
        # set locale to something that doesn't use '.' for the decimal point
        # float must not accept the locale specific decimal point but
        # it still has to accept the normal python syntax
        import locale

        if not locale.localeconv()["decimal_point"] == ",":
            self.skipTest('decimal_point is not ","')

        self.assertEqual(float("  3.14  "), 3.14)
        self.assertEqual(float("+3.14  "), 3.14)
        self.assertEqual(float("-3.14  "), -3.14)
        self.assertEqual(float(".14  "), 0.14)
        self.assertEqual(float("3.  "), 3.0)
        self.assertEqual(float("3.e3  "), 3000.0)
        self.assertEqual(float("3.2e3  "), 3200.0)
        self.assertEqual(float("2.5e-1  "), 0.25)
        self.assertEqual(float("5e-1"), 0.5)
        self.assertRaises(ValueError, float, "  3,14  ")
        self.assertRaises(ValueError, float, "  +3,14  ")
        self.assertRaises(ValueError, float, "  -3,14  ")
        self.assertRaises(ValueError, float, "  0x3.1  ")
        self.assertRaises(ValueError, float, "  -0x3.p-1  ")
        self.assertRaises(ValueError, float, "  +0x3.p-1  ")
        self.assertEqual(float("  25.e-1  "), 2.5)
        self.assertAlmostEqual(float("  .25e-1  "), 0.025)

    def test_floatconversion(self) -> None:  # noqa: C901
        # Make sure that calls to __float__() work properly
        class Foo1(object):
            def __float__(self) -> builtins.float:
                return 42.0

        class Foo2(builtins.float):
            def __float__(self) -> builtins.float:
                return 42.0

        class Foo3(builtins.float):
            def __new__(cls, value: builtins.float = 0.0) -> "Foo3":
                return builtins.float.__new__(cls, 2 * value)

            def __float__(self) -> builtins.float:
                return self

        class Foo4(builtins.float):
            def __float__(self) -> int:
                return 42

        # Issue 5759: __float__ not called on str subclasses (though it is on
        # unicode subclasses).
        class FooStr(str):
            def __float__(self) -> builtins.float:
                return float(str(self)) + 1

        self.assertEqual(float(Foo1()), 42.0)
        self.assertEqual(float(Foo2()), 42.0)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(float(Foo3(21)), 42.0)
        self.assertEqual(float(FooStr("8")), 9.0)
        self.assertRaises(TypeError, float, Foo4(42))

        class Foo5:
            def __float__(self) -> str:
                return ""

        self.assertRaises(TypeError, time.sleep, Foo5())

        # Issue #24731
        class F:
            def __float__(self) -> builtins.float:
                return OtherFloatSubclass(42.0)

        with self.assertWarns(DeprecationWarning):
            self.assertEqual(float(F()), 42.0)
        with self.assertWarns(DeprecationWarning):
            self.assertIs(type(float(F())), builtins.float)
        with self.assertWarns(DeprecationWarning):
            self.assertEqual(FloatSubclass(F()), 42.0)
        with self.assertWarns(DeprecationWarning):
            self.assertIs(type(FloatSubclass(F())), FloatSubclass)

        if sys.version_info >= (3, 8):

            class MyIndex:
                def __init__(self, value: int) -> None:
                    self.value = value

                def __index__(self) -> int:
                    return self.value

            self.assertEqual(float(MyIndex(42)), 42.0)
            self.assertRaises(OverflowError, float, MyIndex(2**2000))

            class MyInt:
                def __int__(self) -> int:
                    return 42

            self.assertRaises(TypeError, float, MyInt())

    def test_keyword_args(self) -> None:
        with self.assertRaisesRegex(TypeError, "keyword argument"):
            float(x="3.14")


# Beginning with Python 2.6 float has cross platform compatible
# ways to create and represent inf and nan
class InfNanTest(unittest.TestCase):
    def test_inf_from_str(self) -> None:
        self.assertTrue(isinf(float("inf")))
        self.assertTrue(isinf(float("+inf")))
        self.assertTrue(isinf(float("-inf")))
        self.assertTrue(isinf(float("infinity")))
        self.assertTrue(isinf(float("+infinity")))
        self.assertTrue(isinf(float("-infinity")))

        self.assertEqual(repr(float("inf")), "inf")
        self.assertEqual(repr(float("+inf")), "inf")
        self.assertEqual(repr(float("-inf")), "-inf")
        self.assertEqual(repr(float("infinity")), "inf")
        self.assertEqual(repr(float("+infinity")), "inf")
        self.assertEqual(repr(float("-infinity")), "-inf")

        self.assertEqual(repr(float("INF")), "inf")
        self.assertEqual(repr(float("+Inf")), "inf")
        self.assertEqual(repr(float("-iNF")), "-inf")
        self.assertEqual(repr(float("Infinity")), "inf")
        self.assertEqual(repr(float("+iNfInItY")), "inf")
        self.assertEqual(repr(float("-INFINITY")), "-inf")

        self.assertEqual(str(float("inf")), "inf")
        self.assertEqual(str(float("+inf")), "inf")
        self.assertEqual(str(float("-inf")), "-inf")
        self.assertEqual(str(float("infinity")), "inf")
        self.assertEqual(str(float("+infinity")), "inf")
        self.assertEqual(str(float("-infinity")), "-inf")

        self.assertRaises(ValueError, float, "info")
        self.assertRaises(ValueError, float, "+info")
        self.assertRaises(ValueError, float, "-info")
        self.assertRaises(ValueError, float, "in")
        self.assertRaises(ValueError, float, "+in")
        self.assertRaises(ValueError, float, "-in")
        self.assertRaises(ValueError, float, "infinit")
        self.assertRaises(ValueError, float, "+Infin")
        self.assertRaises(ValueError, float, "-INFI")
        self.assertRaises(ValueError, float, "infinitys")

        self.assertRaises(ValueError, float, "++Inf")
        self.assertRaises(ValueError, float, "-+inf")
        self.assertRaises(ValueError, float, "+-infinity")
        self.assertRaises(ValueError, float, "--Infinity")

    def test_inf_as_str(self) -> None:
        self.assertEqual(repr(1e300 * 1e300), "inf")
        self.assertEqual(repr(-1e300 * 1e300), "-inf")

        self.assertEqual(str(1e300 * 1e300), "inf")
        self.assertEqual(str(-1e300 * 1e300), "-inf")

    def test_nan_from_str(self) -> None:
        self.assertTrue(isnan(float("nan")))
        self.assertTrue(isnan(float("+nan")))
        self.assertTrue(isnan(float("-nan")))

        self.assertEqual(repr(float("nan")), "nan")
        self.assertEqual(repr(float("+nan")), "nan")
        self.assertEqual(repr(float("-nan")), "nan")

        self.assertEqual(repr(float("NAN")), "nan")
        self.assertEqual(repr(float("+NAn")), "nan")
        self.assertEqual(repr(float("-NaN")), "nan")

        self.assertEqual(str(float("nan")), "nan")
        self.assertEqual(str(float("+nan")), "nan")
        self.assertEqual(str(float("-nan")), "nan")

        self.assertRaises(ValueError, float, "nana")
        self.assertRaises(ValueError, float, "+nana")
        self.assertRaises(ValueError, float, "-nana")
        self.assertRaises(ValueError, float, "na")
        self.assertRaises(ValueError, float, "+na")
        self.assertRaises(ValueError, float, "-na")

        self.assertRaises(ValueError, float, "++nan")
        self.assertRaises(ValueError, float, "-+NAN")
        self.assertRaises(ValueError, float, "+-NaN")
        self.assertRaises(ValueError, float, "--nAn")

    def test_nan_as_str(self) -> None:
        self.assertEqual(repr(1e300 * 1e300 * 0), "nan")
        self.assertEqual(repr(-1e300 * 1e300 * 0), "nan")

        self.assertEqual(str(1e300 * 1e300 * 0), "nan")
        self.assertEqual(str(-1e300 * 1e300 * 0), "nan")

    def test_inf_signs(self) -> None:
        self.assertEqual(copysign(1.0, float("inf")), 1.0)
        self.assertEqual(copysign(1.0, float("-inf")), -1.0)

    @unittest.skipUnless(
        getattr(sys, "float_repr_style", "") == "short",
        "applies only when using short float repr style",
    )
    def test_nan_signs(self) -> None:
        # When using the dtoa.c code, the sign of float('nan') should
        # be predictable.
        self.assertEqual(copysign(1.0, float("nan")), 1.0)
        self.assertEqual(copysign(1.0, float("-nan")), -1.0)


if __name__ == "__main__":
    unittest.main()



"""
Copies doctest.py from the stdlib to the current directory,
and modifies it so that

  a) It will load "*.so" files as modules just like a "*.py" file
  b) It recognizes functions defined in "*.so" files
  c) Remove the configuration extension from the "*.so" files (on Python3)

With these enhancements, doctest can be run on a C python extension module.
"""
from __future__ import print_function

import doctest
import inspect
import sys

# Get the source file location
dt_location = inspect.getsourcefile(doctest)

# Read the module into memory
with open(dt_location) as fl:
    doctest_str = fl.read()

# Let's add the glob module.
# Also define a function to detect if we could import this module name.
doctest_str = doctest_str.replace(
    "import __future__",
    "import __future__\nimport glob\n"
    "def can_import_module(name):\n"
    "    try:\n"
    "        __import__(name)\n"
    "    except ImportError:\n"
    "        return False\n"
    "    else:\n"
    "        return True\n",
)

# Add a search for the .so extension when inspecting the input files
# so that extension modules will be loaded properly.
doctest_str = doctest_str.replace(
    'if filename.endswith(".py"):',
    'if filename.endswith((".py", ".so", ".pyd")) or can_import_module(filename):',
)

# inspect.isfunction does not work for functions written in C,
# so we have to replace that with an inspect.isbuiltin check when
# looking for functions with docstrings.
doctest_str = doctest_str.replace(
    "if ((inspect.isfunction(val) or inspect.isclass(val)) and",
    "if ((inspect.isfunction(val) or inspect.isbuiltin(val) "
    "or inspect.isclass(val)) and",
)

# Replace the configuration extension with nothing on Python3
if sys.version[0] == "3":
    doctest_str = doctest_str.replace(
        "m = __import__(filename[:-3])",
        'm = __import__(filename[:-3] if filename.endswith(".py") '
        'else filename.replace(get_config_var("EXT_SUFFIX"), ""))',
    )
    # We need to import the get_config_var variable.
    doctest_str = doctest_str.replace(
        "def _test():", "from sysconfig import get_config_var\ndef _test():"
    )
else:
    doctest_str = doctest_str.replace(
        "m = __import__(filename[:-3])",
        "m = __import__(filename[:-3] "
        'if filename.endswith((".py", ".so", ".pyd")) else filename)',
    )

# To silence mypy
doctest_str = "# type: ignore\n" + doctest_str

# Open up the new output file and write the modified input to it.
with open("doctest.py", "w") as fl:
    print(doctest_str, file=fl, end="")



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import glob
import os
import sys

from setuptools import Extension, find_packages, setup


# Compilation arguments are platform-dependent
if sys.platform == "win32":
    compile_args = [
        "/std:c++17",
        "/W4",  # Equivalent to -Wall -Wpedantic on GCC/Clang
        "/wd4100",  # don't warn about unreferenced formal parameters
        "/wd4127",  # don't warn about constant conditional expressions
    ]
    if "FN_DEBUG" in os.environ or "FN_COV" in os.environ:
        compile_args.append("/Od")
        compile_args.append("/Z7")
else:
    compile_args = [
        "-std=c++17",
        "-Wall",
        "-Weffc++",
        "-Wpedantic",
    ]
    if sys.platform == "darwin":
        compile_args.append("-mmacosx-version-min=10.13")
    if "FN_DEBUG" in os.environ or "FN_COV" in os.environ:
        compile_args.append("-Og")
        compile_args.append("-g")


ext = [
    Extension(
        "fastnumbers.fastnumbers",
        sorted(glob.glob("src/cpp/*.cpp")),
        include_dirs=[os.path.abspath(os.path.join("include"))],
        extra_compile_args=compile_args,
        extra_link_args=["-lm"],
    )
]

# Define how to build the extension module.
# All other data is in the setup.cfg file.
setup(
    name="fastnumbers",
    version="5.0.0",
    python_requires=">=3.7",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    package_data={"fastnumbers": ["py.typed", "*.pyi"]},
    zip_safe=False,
    ext_modules=ext,
)
